from __future__ import annotations

from datetime import timedelta
from typing import Any, Callable

import pyarrow as pa

from chalk._lsp.error_builder import get_resolver_error_builder
from chalk.features._embedding.cohere import CohereProvider
from chalk.features._embedding.embedding_provider import EmbeddingProvider
from chalk.features._embedding.openai import OpenAIProvider
from chalk.features._embedding.sentence_transformer import SentenceTransformerProvider
from chalk.features._embedding.vertexai import VertexAIProvider
from chalk.features.dataframe import DataFrame
from chalk.features.feature_field import Feature
from chalk.features.feature_set import Features
from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
from chalk.features.resolver import RESOLVER_REGISTRY, OnlineResolver
from chalk.features.underscore import Underscore, UnderscoreFunction
from chalk.serialization.parsed_annotation import ParsedAnnotation
from chalk.utils.collections import ensure_tuple

SUPPORTED_LOCAL_MODELS = {
    "all-MiniLM-L6-v2",  # https://huggingface.co/sentence-transformers/all-MiniLM-L6-v2
    "sample-bert",  # For internal Chalk use
}


def _get_provider(provider: str, model: str, dimensions: int | None = None) -> EmbeddingProvider:
    """Factory function to get an EmbeddingProvider"""
    if provider == "openai":
        return OpenAIProvider(model, dimensions)
    elif provider == "cohere":
        return CohereProvider(model, dimensions)
    elif provider == "vertexai":
        return VertexAIProvider(model, dimensions)
    elif provider == "sentence-transformers":
        return SentenceTransformerProvider(model, dimensions)
    raise ValueError(
        f"Unsupported embedding provider: {provider}. The supported providers are ['openai', 'cohere', 'vertexai']."
    )


def embed(
    input: Callable[[], Any],
    model: str | None = None,
    model_path: str | None = None,
    tokenizer_path: str | None = None,
    provider: str | None = None,
    name: str | None = None,
    owner: str | None = None,
    tags: list[str] | None = None,
    max_staleness: str | timedelta | None | ellipsis = ...,
    description: str | None = None,
    version: int | None = None,
    default_version: int = 1,
    etl_offline_to_online: bool | None = None,
    offline_ttl: ellipsis | str | timedelta | None = ...,
    default: Any = None,
    dimensions: int | None = None,
) -> Any:
    """Specify an embedding feature.

    Parameters
    ----------
    input
        The input for the embedding. This argument is callable
        to allow for forward references to features of the same
        class.
    provider
        The AI provider to use for the embedding.
    model
        The model to generate the embedding.
    model_path
        The path to the embedding model file: currently, `onnx` models are supported.
    tokenizer_path
        The path to the tokenizer for the model file. This is only used running an embedding model locally.
    dimensions
        The dimensionality of the embedding. If not specified, the default dimensionality will be used.
        Supported dimensionality varies based on the model.
    owner
        You may also specify which person or group is responsible for a feature.
        The owner tag will be available in Chalk's web portal.
        Alerts that do not otherwise have an owner will be assigned
        to the owner of the monitored feature.
        Read more at https://docs.chalk.ai/docs/feature-discovery#owner
    tags
        Add metadata to a feature for use in filtering, aggregations,
        and visualizations. For example, you can use tags to assign
        features to a team and find all features for a given team.
        Read more at https://docs.chalk.ai/docs/feature-discovery#tags
    max_staleness
        When a feature is expensive or slow to compute, you may wish to cache its value.
        Chalk uses the terminology "maximum staleness" to describe how recently a feature
        value needs to have been computed to be returned without re-running a resolver.
        Read more at https://docs.chalk.ai/docs/feature-caching
    etl_offline_to_online
        When `True`, Chalk copies this feature into the online environment
        when it is computed in offline resolvers.
        Read more at https://docs.chalk.ai/docs/reverse-etl
    version
        The maximum version for a feature. Versioned features can be
        referred to with the `@` operator:

        >>> from chalk.features import Vector, features
        >>> @features
        ... class Document:
        ...     id: str
        ...     content: str
        ...     score: Vector = embed(
        ...         input=lambda: Document.content,
        ...         provider="openai",
        ...         model="text-embedding-ada-002",
        ...         version=2,
        ...     )
        >>> str(Document.content @ 2)
        "document.content@2"

        See more at https://docs.chalk.ai/docs/feature-versions
    default_version
        The default version for a feature. When you reference a
        versioned feature without the `@` operator, you reference
        the `default_version`. Set to `1` by default.

        >>> from chalk.features import Vector, features
        >>> @features
        ... class Document:
        ...     id: str
        ...     content: str
        ...     embedding: Vector = embed(
        ...         input=lambda: Document.content,
        ...         provider="openai",
        ...         model="text-embedding-ada-002",
        ...         version=2,
        ...         default_version=2,
        ...     )
        >>> str(Document.content)
        "document.content"

        See more at https://docs.chalk.ai/docs/feature-versions#default-versions

    Other Parameters
    ----------------
    name
        The name for the feature. By default, the name of a feature is
        the name of the attribute on the class, prefixed with
        the camel-cased name of the class. Note that if you provide an
        explicit name, the namespace, determined by the feature class,
        will still be prepended. See `features` for more details.
    description
        Descriptions are typically provided as comments preceding
        the feature definition. For example, you can document a
        `fraud_score` feature with information about the values
        as follows:

        >>> @features
        ... class Document:
        ...     # 0 to 100 score indicating an identity match.
        ...     embedding: Vector = embed(...)

        You can also specify the description directly with this parameter.
        Read more at https://docs.chalk.ai/docs/feature-discovery#description
    offline_ttl
    default
        The default value to use for this embedding feature when no value is available.
        This can be useful when you want to provide a fallback embedding value.

    Examples
    --------
    >>> from chalk.features import Vector, features
    >>> @features
    ... class Document:
    ...     id: str
    ...     content: str
    ...     embedding: Vector = embed(
    ...         input=lambda: Document.content,
    ...         provider="openai",
    ...         model="text-embedding-ada-002",
    ...     )
    """
    if provider is None:
        if not isinstance(input, Underscore):
            raise TypeError(
                "When using `model_path` and `provider`, the `input` must be an underscore not a lambda expression: `_.content`."
            )

        if model in SUPPORTED_LOCAL_MODELS:
            return UnderscoreFunction("onnx_run_embedding", input, model_name=model)

        if model_path is not None and tokenizer_path is not None:
            return UnderscoreFunction("onnx_run_embedding", input, model_path=model_path, tokenizer_path=tokenizer_path)

        raise ValueError(
            f"If `provider` is None, you must specify either: both a `model_path` and `tokenizer_path`, or one of the supported local embedding models ({', '.join(list(SUPPORTED_LOCAL_MODELS))})."
        )

    if model is None:
        raise ValueError("If provider is set, then model must also be set")

    embedding_provider = _get_provider(provider, model, dimensions)
    # Manually set the dimensions of the Vector when using embedding
    typ = ParsedAnnotation(underlying=embedding_provider.get_vector_class())
    output_feature = Feature(
        name=name,
        owner=owner,
        tags=tags,
        typ=typ,
        max_staleness=max_staleness,
        description=description,
        version=version,
        default_version=default_version,
        etl_offline_to_online=etl_offline_to_online,
        offline_ttl=offline_ttl,
        default=default,
    )
    previous_hook = output_feature.hook

    def hook(features: type[Features]) -> None:
        if previous_hook:
            previous_hook(features)

        def resolver_factory():
            inputs = ensure_tuple(input())
            input_features_or_literals = tuple(
                unwrap_feature(x) if isinstance(x, (Feature, FeatureWrapper)) else x for x in inputs
            )
            input_features: list[Feature] = []
            input_schema: list[pa.DataType] = []
            for x in input_features_or_literals:
                if isinstance(x, Feature):
                    input_features.append(x)
                    input_schema.append(x.converter.pyarrow_dtype)
                    continue
                if not isinstance(x, str):
                    raise TypeError("Embedding function literals must be strings")
                input_schema.append(pa.large_utf8())
            if features.__chalk_primary__ not in input_features:
                assert features.__chalk_primary__ is not None
                input_features.append(features.__chalk_primary__)
            error_str = embedding_provider.validate_input_schema(input_schema)
            if error_str:
                raise ValueError(
                    (
                        f"The first argument of the `embedding` function for feature '{output_feature.root_fqn}' returned an "
                        f"unsupported input schema for embedding model "
                        f"'{embedding_provider.get_provider_name()}/{embedding_provider.get_model_name()}': {error_str}."
                    )
                )

            async def fn(raw_inputs: DataFrame):
                # We need to build the input table by combining the literals with the features
                input_arrays: list[pa.Array | pa.ChunkedArray] = []
                raw_input_table = raw_inputs.to_pyarrow()
                assert features.__chalk_primary__ is not None
                pkeys = raw_input_table.column(features.__chalk_primary__.root_fqn)
                for x in input_features_or_literals:
                    if isinstance(x, Feature):
                        input_arrays.append(raw_input_table.column(x.root_fqn))
                    else:
                        input_arrays.append(pa.nulls(len(raw_input_table), pa.large_utf8()).fill_null(x))
                unified_inputs_table = pa.Table.from_arrays(
                    input_arrays, names=[f"col_{i}" for i in range(len(input_arrays))]
                )
                i = 0
                async for embeddings in embedding_provider.async_generate_embedding(unified_inputs_table):
                    t = pa.Table.from_arrays(
                        [embeddings, pkeys.slice(i, len(embeddings))],
                        [output_feature.root_fqn, features.__chalk_primary__.root_fqn],
                    )
                    i += len(embeddings)

                    yield t

            return OnlineResolver(
                function_definition="",
                filename="",
                fqn=f"__chalk__embedding__resolver__namespace__{output_feature.namespace}__name__{output_feature.name}",
                doc=None,
                inputs=[DataFrame[tuple(input_features)]],
                state=None,
                output=Features[DataFrame[output_feature, features.__chalk_primary__]],
                fn=fn,
                environment=None,
                tags=output_feature.tags,
                machine_type=None,
                default_args=[None],
                owner=output_feature.owner,
                timeout=None,
                cron=None,
                when=None,
                data_sources=None,
                is_sql_file_resolver=False,
                source_line=None,
                lsp_builder=get_resolver_error_builder(fn),
                parse=None,
                resource_hint=None,
                static=False,
                total=False,
                autogenerated=False,
                unique_on=None,
                partitioned_by=None,
                data_lineage=None,
                sql_settings=None,
            )

        RESOLVER_REGISTRY.add_to_deferred_registry(resolver_factory, override=False)

    output_feature.hook = hook

    return output_feature



from __future__ import annotations

import re
from datetime import timedelta
from re import Pattern
from typing import Literal, Mapping, Optional, Tuple, Union

from typing_extensions import TypeAlias

Duration: TypeAlias = Union[str, timedelta, Literal["infinity", "all"]]
"""
Duration is used to describe time periods in natural language.
To specify using natural language, write the count of the unit
you would like, followed by the representation of the unit.

Chalk supports the following units:

 | Signifier | Meaning       |
 | --------- | ------------- |
 | w         | Weeks         |
 | d         | Days          |
 | h         | Hours         |
 | m         | Minutes       |
 | s         | Seconds       |
 | ms        | Milliseconds  |

As well as the special keywords `"infinity"` and `"all"`.

Examples:

| Signifier            | Meaning                           |
| -------------------- | --------------------------------- |
| "10h"                | 10 hours                          |
| "1w 2m"              | 1 week and 2 minutes              |
| "1h 10m 2s"          | 1 hour, 10 minutes, and 2 seconds |
| "infinity" and "all" | Unbounded time duration           |
"""

CronTab: TypeAlias = str
"""
A schedule defined using the Unix-cron
string format (`* * * * *`).
Values are given in the order below:


| Field        | Values |
| ------------ | ------ |
| Minute       | 0-59   |
| Hour         | 0-23   |
| Day of Month | 1-31   |
| Month        | 1-12   |
| Day of Week  | 0-6    |
"""

ScheduleOptions: TypeAlias = Optional[Union[CronTab, Duration, Literal[True]]]
"""The schedule on which to run a resolver.

One of:
- `CronTab`: A Unix-cron string, e.g. `"* * * * *"`.
- `Duration`: A Chalk Duration, e.g. `"2h30m"`.
"""


_kwarg_to_regex: Mapping[str, Pattern] = {
    timedelta_kwarg: re.compile(rf"(([0-9]+?){abbreviation}(?![a-z]))")
    for timedelta_kwarg, abbreviation in dict(
        weeks="w",
        days="d",
        hours="h",
        minutes="m",
        seconds="s",
        milliseconds="ms",
    ).items()
}


def _parse(s: str, regex: Pattern) -> Tuple[int, str]:
    matched = regex.search(s)
    if matched is None:
        return 0, ""

    return int(matched.groups()[1]), matched.groups()[0]


CHALK_MAX_TIMEDELTA = timedelta(days=100 * 365)
"""The maximum duration supported that can be represented in an int64_t in nanoseconds, rounded down to a nice number."""


def parse_chalk_duration_s(s: str | timedelta | int | Literal["infinity"]) -> int:
    return int(parse_chalk_duration(s).total_seconds())


def parse_chalk_duration(s: str | timedelta | int | Literal["infinity", "all"]) -> timedelta:
    """Parses any form of Chalk duration into a timedelta.

    If conversion fails, a value error is raised with a friendly error message the as the only arg.
    """
    if isinstance(s, timedelta):
        return s
    elif isinstance(s, int):
        return CHALK_MAX_TIMEDELTA if s >= CHALK_MAX_TIMEDELTA.total_seconds() else timedelta(seconds=s)
    elif s == "infinity" or s == "all":
        return CHALK_MAX_TIMEDELTA

    if not isinstance(s, str):  # pyright: ignore[reportUnnecessaryIsInstance]
        raise ValueError(
            f"Expected a string, timedelta, or integer, but got `{s}`. "
            + "Please use a valid duration, like '10m', '1h', or '1h30m'. "
            + "Read more at https://docs.chalk.ai/api-docs#Duration"
        )

    negative = s.startswith("-")
    s = s.removeprefix("-")

    parsed_values: dict[str, tuple[int, str]] = {k: _parse(s, unit) for k, unit in _kwarg_to_regex.items()}

    # Check for remaining unparsed input
    timedelta_constructor: dict[str, int] = {}
    remainder = s
    for k, v in parsed_values.items():
        remainder = remainder.replace(v[1], "", 1)
        timedelta_constructor[k] = v[0]
    remainder = remainder.strip()

    if remainder != "":
        raise ValueError(
            (
                f"The duration '{s}' contained a component '{remainder}' that could not be parsed. "
                "Please use a valid duration, like '10m', '1h', or '1h30m'. "
                "Read more at https://docs.chalk.ai/api-docs#Duration"
            )
        )

    td = timedelta(**timedelta_constructor)
    return -td if negative else td


def timedelta_to_duration(td: timedelta | int) -> str:
    if isinstance(td, int):
        td = timedelta(seconds=td)

    if td >= CHALK_MAX_TIMEDELTA:
        return "infinity"
    total_seconds = td.total_seconds()
    negative = total_seconds < 0
    total_seconds = abs(total_seconds)
    seconds = int(total_seconds)
    milliseconds = int((total_seconds - seconds) * 1000)
    days, remainder = divmod(seconds, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)
    duration = []
    if days > 0:
        duration.append(f"{days}d")
    if hours > 0:
        duration.append(f"{hours}h")
    if minutes > 0:
        duration.append(f"{minutes}m")
    if seconds > 0:
        duration.append(f"{seconds}s")
    if milliseconds > 0:
        duration.append(f"{milliseconds}ms")
    return ("-" if negative else "") + "".join(duration)



from chalk._gen.chalk.artifacts.v1 import cdc_pb2 as _cdc_pb2
from chalk._gen.chalk.artifacts.v1 import chart_pb2 as _chart_pb2
from chalk._gen.chalk.artifacts.v1 import cron_query_pb2 as _cron_query_pb2
from chalk._gen.chalk.artifacts.v1 import export_pb2 as _export_pb2
from chalk._gen.chalk.graph.v1 import graph_pb2 as _graph_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class DeploymentArtifacts(_message.Message):
    __slots__ = ("graph", "crons", "charts", "cdc_sources", "config", "chalkpy")
    GRAPH_FIELD_NUMBER: _ClassVar[int]
    CRONS_FIELD_NUMBER: _ClassVar[int]
    CHARTS_FIELD_NUMBER: _ClassVar[int]
    CDC_SOURCES_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    CHALKPY_FIELD_NUMBER: _ClassVar[int]
    graph: _graph_pb2.Graph
    crons: _containers.RepeatedCompositeFieldContainer[_cron_query_pb2.CronQuery]
    charts: _containers.RepeatedCompositeFieldContainer[_chart_pb2.Chart]
    cdc_sources: _containers.RepeatedCompositeFieldContainer[_cdc_pb2.CDCSource]
    config: _export_pb2.ProjectSettings
    chalkpy: _export_pb2.ChalkpyInfo
    def __init__(
        self,
        graph: _Optional[_Union[_graph_pb2.Graph, _Mapping]] = ...,
        crons: _Optional[_Iterable[_Union[_cron_query_pb2.CronQuery, _Mapping]]] = ...,
        charts: _Optional[_Iterable[_Union[_chart_pb2.Chart, _Mapping]]] = ...,
        cdc_sources: _Optional[_Iterable[_Union[_cdc_pb2.CDCSource, _Mapping]]] = ...,
        config: _Optional[_Union[_export_pb2.ProjectSettings, _Mapping]] = ...,
        chalkpy: _Optional[_Union[_export_pb2.ChalkpyInfo, _Mapping]] = ...,
    ) -> None: ...



from typing import TYPE_CHECKING, Any, List, Literal, Mapping, Optional, Union

from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sink._models import SinkIntegrationProtocol
from chalk.streams.base import StreamSource
from chalk.utils.duration import Duration
from chalk.utils.string import comma_join, comma_whitespace_split

if TYPE_CHECKING:
    from pydantic import BaseModel, Field
else:
    try:
        from pydantic.v1 import BaseModel, Field
    except ImportError:
        from pydantic import BaseModel, Field


_KAFKA_BOOTSTRAP_SERVER_NAME = "KAFKA_BOOTSTRAP_SERVER"
_KAFKA_TOPIC_NAME = "KAFKA_TOPIC"
_KAFKA_SSL_KEYSTORE_LOCATION_NAME = "KAFKA_SSL_KEYSTORE_LOCATION"
_KAFKA_SSL_CA_FILE_NAME = "KAFKA_SSL_CA_FILE"
_KAFKA_CLIENT_ID_PREFIX_NAME = "KAFKA_CLIENT_ID_PREFIX"
_KAFKA_GROUP_ID_PREFIX_NAME = "KAFKA_GROUP_ID_PREFIX"
_KAFKA_SECURITY_PROTOCOL_NAME = "KAFKA_SECURITY_PROTOCOL"
_KAFKA_SASL_MECHANISM_NAME = "KAFKA_SASL_MECHANISM"
_KAFKA_SASL_USERNAME_NAME = "KAFKA_SASL_USERNAME"
_KAFKA_SASL_PASSWORD_NAME = "KAFKA_SASL_PASSWORD"


class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True):
    bootstrap_server: Optional[Union[str, List[str]]] = None
    """The URL of one of your Kafka brokers from which to fetch initial metadata about your Kafka cluster"""

    topic: Optional[str] = None
    """The name of the topic to subscribe to."""

    ssl_keystore_location: Optional[str] = None
    """
    An S3 or GCS URI that points to the keystore file that should be
    used for brokers. You must configure the appropriate AWS or
    GCP integration in order for Chalk to be able to access these
    files.
    """

    ssl_ca_file: Optional[str] = None
    """
    An S3 or GCS URI that points to the certificate authority file that should be
    used to verify broker certificates. You must configure the appropriate AWS or
    GCP integration in order for Chalk to be able to access these files.
    """

    client_id_prefix: str = "chalk/"
    group_id_prefix: str = "chalk/"

    security_protocol: Literal["PLAINTEXT", "SSL", "SASL_PLAINTEXT", "SASL_SSL"] = "PLAINTEXT"
    """
    Protocol used to communicate with brokers.
    Valid values are `"PLAINTEXT"`, `"SSL"`, `"SASL_PLAINTEXT"`, and `"SASL_SSL"`.
    Defaults to `"PLAINTEXT"`.
    """

    sasl_mechanism: Literal["PLAIN", "GSSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512", "OAUTHBEARER"] = "PLAIN"
    """
    Authentication mechanism when `security_protocol`
    is configured for SASL_PLAINTEXT or SASL_SSL.
    Valid values are `"PLAIN"`, `"GSSAPI"`, `"SCRAM-SHA-256"`, `"SCRAM-SHA-512"`, `"OAUTHBEARER"`.
    Defaults to `"PLAIN"`.
    """

    sasl_username: Optional[str] = None
    """
    Username for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
    """

    sasl_password: Optional[str] = Field(default=None, repr=False)
    """
    Password for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
    """

    name: Optional[str] = None
    """
    The name of the integration, as configured in your Chalk Dashboard.
    """

    late_arrival_deadline: Duration = "infinity"
    """
    Messages older than this deadline will not be processed.
    """

    dead_letter_queue_topic: Optional[str] = None
    """
    Kafka topic to send messages when message processing fails
    """

    def __init__(
        self,
        *,
        bootstrap_server: Optional[Union[str, List[str]]] = None,
        topic: Optional[str] = None,
        ssl_keystore_location: Optional[str] = None,
        ssl_ca_file: Optional[str] = None,
        client_id_prefix: Optional[str] = None,
        group_id_prefix: Optional[str] = None,
        security_protocol: Optional[str] = None,
        sasl_mechanism: Optional[Literal["PLAIN", "GSSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512", "OAUTHBEARER"]] = None,
        sasl_username: Optional[str] = None,
        sasl_password: Optional[str] = None,
        name: Optional[str] = None,
        late_arrival_deadline: Duration = "infinity",
        dead_letter_queue_topic: Optional[str] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        super(KafkaSource, self).__init__(
            bootstrap_server=bootstrap_server
            or load_integration_variable(
                name=_KAFKA_BOOTSTRAP_SERVER_NAME,
                integration_name=name,
                parser=comma_whitespace_split,
                override=integration_variable_override,
            ),
            topic=topic
            or load_integration_variable(
                name=_KAFKA_TOPIC_NAME, integration_name=name, override=integration_variable_override
            ),
            ssl_keystore_location=ssl_keystore_location
            or load_integration_variable(
                name=_KAFKA_SSL_KEYSTORE_LOCATION_NAME, integration_name=name, override=integration_variable_override
            ),
            client_id_prefix=client_id_prefix
            or load_integration_variable(
                name=_KAFKA_CLIENT_ID_PREFIX_NAME, integration_name=name, override=integration_variable_override
            )
            or KafkaSource.__fields__["client_id_prefix"].default,
            group_id_prefix=group_id_prefix
            or load_integration_variable(
                name=_KAFKA_GROUP_ID_PREFIX_NAME, integration_name=name, override=integration_variable_override
            )
            or KafkaSource.__fields__["group_id_prefix"].default,
            security_protocol=security_protocol
            or load_integration_variable(
                name=_KAFKA_SECURITY_PROTOCOL_NAME, integration_name=name, override=integration_variable_override
            )
            or KafkaSource.__fields__["security_protocol"].default,
            sasl_mechanism=sasl_mechanism
            or load_integration_variable(
                name=_KAFKA_SASL_MECHANISM_NAME, integration_name=name, override=integration_variable_override
            )
            or KafkaSource.__fields__["sasl_mechanism"].default,
            sasl_username=sasl_username
            or load_integration_variable(
                name=_KAFKA_SASL_USERNAME_NAME, integration_name=name, override=integration_variable_override
            ),
            sasl_password=sasl_password
            or load_integration_variable(
                name=_KAFKA_SASL_PASSWORD_NAME, integration_name=name, override=integration_variable_override
            ),
            name=name,
            late_arrival_deadline=late_arrival_deadline,
            dead_letter_queue_topic=dead_letter_queue_topic,
            ssl_ca_file=ssl_ca_file
            or load_integration_variable(
                name=_KAFKA_SSL_CA_FILE_NAME, integration_name=name, override=integration_variable_override
            ),
        )
        self.registry.append(self)

    def config_to_json(self) -> Any:
        return self.json()

    @property
    def streaming_type(self) -> str:
        return "kafka"

    @property
    def dlq_name(self) -> Union[str, None]:
        return self.dead_letter_queue_topic

    @property
    def stream_or_topic_name(self):
        assert self.topic is not None
        return self.topic

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(
                    _KAFKA_BOOTSTRAP_SERVER_NAME, self.name, self.bootstrap_server, serializer=comma_join
                ),
                create_integration_variable(_KAFKA_TOPIC_NAME, self.name, self.topic),
                create_integration_variable(_KAFKA_SSL_KEYSTORE_LOCATION_NAME, self.name, self.ssl_keystore_location),
                create_integration_variable(_KAFKA_SSL_CA_FILE_NAME, self.name, self.ssl_ca_file),
                create_integration_variable(_KAFKA_CLIENT_ID_PREFIX_NAME, self.name, self.client_id_prefix),
                create_integration_variable(_KAFKA_GROUP_ID_PREFIX_NAME, self.name, self.group_id_prefix),
                create_integration_variable(_KAFKA_SECURITY_PROTOCOL_NAME, self.name, self.security_protocol),
                create_integration_variable(_KAFKA_SASL_MECHANISM_NAME, self.name, self.sasl_mechanism),
                create_integration_variable(_KAFKA_SASL_USERNAME_NAME, self.name, self.sasl_username),
                create_integration_variable(_KAFKA_SASL_PASSWORD_NAME, self.name, self.sasl_password),
            ]
            if v is not None
        }



from chalk._gen.chalk.auth.v1 import featurepermission_pb2 as _featurepermission_pb2
from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class DisplayPermission(_message.Message):
    __slots__ = ("name", "permission")
    NAME_FIELD_NUMBER: _ClassVar[int]
    PERMISSION_FIELD_NUMBER: _ClassVar[int]
    name: str
    permission: _permissions_pb2.Permission
    def __init__(
        self, name: _Optional[str] = ..., permission: _Optional[_Union[_permissions_pb2.Permission, str]] = ...
    ) -> None: ...

class DisplayEnvironmentPermissions(_message.Message):
    __slots__ = ("environment_id", "environment_name", "project_id", "project_name", "permissions")
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    PROJECT_ID_FIELD_NUMBER: _ClassVar[int]
    PROJECT_NAME_FIELD_NUMBER: _ClassVar[int]
    PERMISSIONS_FIELD_NUMBER: _ClassVar[int]
    environment_id: str
    environment_name: str
    project_id: str
    project_name: str
    permissions: _containers.RepeatedCompositeFieldContainer[DisplayPermission]
    def __init__(
        self,
        environment_id: _Optional[str] = ...,
        environment_name: _Optional[str] = ...,
        project_id: _Optional[str] = ...,
        project_name: _Optional[str] = ...,
        permissions: _Optional[_Iterable[_Union[DisplayPermission, _Mapping]]] = ...,
    ) -> None: ...

class DisplayUserAgent(_message.Message):
    __slots__ = (
        "client_id",
        "user_id",
        "team_id",
        "permissions_by_environment",
        "impersonated",
        "name",
        "email",
        "team_name",
        "customer_claims",
    )
    CLIENT_ID_FIELD_NUMBER: _ClassVar[int]
    USER_ID_FIELD_NUMBER: _ClassVar[int]
    TEAM_ID_FIELD_NUMBER: _ClassVar[int]
    PERMISSIONS_BY_ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    IMPERSONATED_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    TEAM_NAME_FIELD_NUMBER: _ClassVar[int]
    CUSTOMER_CLAIMS_FIELD_NUMBER: _ClassVar[int]
    client_id: str
    user_id: str
    team_id: str
    permissions_by_environment: _containers.RepeatedCompositeFieldContainer[DisplayEnvironmentPermissions]
    impersonated: bool
    name: str
    email: str
    team_name: str
    customer_claims: _containers.RepeatedCompositeFieldContainer[DisplayCustomClaim]
    def __init__(
        self,
        client_id: _Optional[str] = ...,
        user_id: _Optional[str] = ...,
        team_id: _Optional[str] = ...,
        permissions_by_environment: _Optional[_Iterable[_Union[DisplayEnvironmentPermissions, _Mapping]]] = ...,
        impersonated: bool = ...,
        name: _Optional[str] = ...,
        email: _Optional[str] = ...,
        team_name: _Optional[str] = ...,
        customer_claims: _Optional[_Iterable[_Union[DisplayCustomClaim, _Mapping]]] = ...,
    ) -> None: ...

class DisplayCustomClaim(_message.Message):
    __slots__ = ("key", "values")
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    key: str
    values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, key: _Optional[str] = ..., values: _Optional[_Iterable[str]] = ...) -> None: ...

class DisplayServiceTokenAgent(_message.Message):
    __slots__ = (
        "id",
        "client_id",
        "team_id",
        "project_id",
        "environment_id",
        "permissions",
        "name",
        "team_name",
        "project_name",
        "environment_name",
        "customer_claims",
        "feature_permissions",
        "created_at",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    CLIENT_ID_FIELD_NUMBER: _ClassVar[int]
    TEAM_ID_FIELD_NUMBER: _ClassVar[int]
    PROJECT_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    PERMISSIONS_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    TEAM_NAME_FIELD_NUMBER: _ClassVar[int]
    PROJECT_NAME_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    CUSTOMER_CLAIMS_FIELD_NUMBER: _ClassVar[int]
    FEATURE_PERMISSIONS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    id: str
    client_id: str
    team_id: str
    project_id: str
    environment_id: str
    permissions: _containers.RepeatedCompositeFieldContainer[DisplayPermission]
    name: str
    team_name: str
    project_name: str
    environment_name: str
    customer_claims: _containers.RepeatedCompositeFieldContainer[DisplayCustomClaim]
    feature_permissions: _featurepermission_pb2.FeaturePermissions
    created_at: _timestamp_pb2.Timestamp
    def __init__(
        self,
        id: _Optional[str] = ...,
        client_id: _Optional[str] = ...,
        team_id: _Optional[str] = ...,
        project_id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        permissions: _Optional[_Iterable[_Union[DisplayPermission, _Mapping]]] = ...,
        name: _Optional[str] = ...,
        team_name: _Optional[str] = ...,
        project_name: _Optional[str] = ...,
        environment_name: _Optional[str] = ...,
        customer_claims: _Optional[_Iterable[_Union[DisplayCustomClaim, _Mapping]]] = ...,
        feature_permissions: _Optional[_Union[_featurepermission_pb2.FeaturePermissions, _Mapping]] = ...,
        created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
    ) -> None: ...

class DisplayEngineAgent(_message.Message):
    __slots__ = (
        "id",
        "team_id",
        "project_id",
        "environment_id",
        "impersonated",
        "team_name",
        "project_name",
        "environment_name",
    )
    ID_FIELD_NUMBER: _ClassVar[int]
    TEAM_ID_FIELD_NUMBER: _ClassVar[int]
    PROJECT_ID_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_ID_FIELD_NUMBER: _ClassVar[int]
    IMPERSONATED_FIELD_NUMBER: _ClassVar[int]
    TEAM_NAME_FIELD_NUMBER: _ClassVar[int]
    PROJECT_NAME_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    id: str
    team_id: str
    project_id: str
    environment_id: str
    impersonated: bool
    team_name: str
    project_name: str
    environment_name: str
    def __init__(
        self,
        id: _Optional[str] = ...,
        team_id: _Optional[str] = ...,
        project_id: _Optional[str] = ...,
        environment_id: _Optional[str] = ...,
        impersonated: bool = ...,
        team_name: _Optional[str] = ...,
        project_name: _Optional[str] = ...,
        environment_name: _Optional[str] = ...,
    ) -> None: ...

class DisplayTenantAgent(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DisplayMetadataServiceAgent(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class DisplayAgent(_message.Message):
    __slots__ = ("user_agent", "service_token_agent", "engine_agent", "tenant_agent", "metadata_service_agent")
    USER_AGENT_FIELD_NUMBER: _ClassVar[int]
    SERVICE_TOKEN_AGENT_FIELD_NUMBER: _ClassVar[int]
    ENGINE_AGENT_FIELD_NUMBER: _ClassVar[int]
    TENANT_AGENT_FIELD_NUMBER: _ClassVar[int]
    METADATA_SERVICE_AGENT_FIELD_NUMBER: _ClassVar[int]
    user_agent: DisplayUserAgent
    service_token_agent: DisplayServiceTokenAgent
    engine_agent: DisplayEngineAgent
    tenant_agent: DisplayTenantAgent
    metadata_service_agent: DisplayMetadataServiceAgent
    def __init__(
        self,
        user_agent: _Optional[_Union[DisplayUserAgent, _Mapping]] = ...,
        service_token_agent: _Optional[_Union[DisplayServiceTokenAgent, _Mapping]] = ...,
        engine_agent: _Optional[_Union[DisplayEngineAgent, _Mapping]] = ...,
        tenant_agent: _Optional[_Union[DisplayTenantAgent, _Mapping]] = ...,
        metadata_service_agent: _Optional[_Union[DisplayMetadataServiceAgent, _Mapping]] = ...,
    ) -> None: ...



from __future__ import annotations

import inspect
import os
from pathlib import Path
from typing import List, Optional


def get_classpath(x: object) -> Path | None:
    module = inspect.getmodule(x)
    if module is None:
        return
    filepath = getattr(module, "__file__", None)
    if filepath is None:
        return
    return Path(os.path.abspath(filepath))


def get_classpath_or_name(x: object) -> str | None:
    classpath = get_classpath(x)
    if classpath is not None:
        return str(classpath)
    return getattr(x, "__module__", None)


def get_directory_root() -> Optional[Path]:
    current = Path(os.path.dirname(os.path.abspath("dummy.txt")))
    while True:
        if any((current / f).exists() for f in ("chalk.yaml", "chalk.yml")):
            return current
        if Path(os.path.dirname(current)) == current:
            # This is '/'
            return None
        current = current.parent


def search_recursively_for_file(base: Path, filename: str) -> List[Path]:
    ans = []
    assert base.is_dir()
    while True:
        filepath = base / filename
        if filepath.exists():
            ans.append(filepath)
        parent = base.parent
        if parent == base:
            return ans
        base = parent



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

from abc import (
    ABCMeta,
    abstractmethod,
)
from chalk._gen.chalk.server.v1.topicpush_pb2 import (
    CreateJobRequest,
    CreateJobResponse,
    DeleteJobRequest,
    DeleteJobResponse,
    GetJobByNameRequest,
    GetJobByNameResponse,
    ListJobsRequest,
    ListJobsResponse,
    UpdateJobRequest,
    UpdateJobResponse,
)
from grpc import (
    Channel,
    Server,
    ServicerContext,
    UnaryUnaryMultiCallable,
)

class TopicPushServiceStub:
    def __init__(self, channel: Channel) -> None: ...
    ListJobs: UnaryUnaryMultiCallable[
        ListJobsRequest,
        ListJobsResponse,
    ]
    CreateJob: UnaryUnaryMultiCallable[
        CreateJobRequest,
        CreateJobResponse,
    ]
    UpdateJob: UnaryUnaryMultiCallable[
        UpdateJobRequest,
        UpdateJobResponse,
    ]
    DeleteJob: UnaryUnaryMultiCallable[
        DeleteJobRequest,
        DeleteJobResponse,
    ]
    GetJobByName: UnaryUnaryMultiCallable[
        GetJobByNameRequest,
        GetJobByNameResponse,
    ]

class TopicPushServiceServicer(metaclass=ABCMeta):
    @abstractmethod
    def ListJobs(
        self,
        request: ListJobsRequest,
        context: ServicerContext,
    ) -> ListJobsResponse: ...
    @abstractmethod
    def CreateJob(
        self,
        request: CreateJobRequest,
        context: ServicerContext,
    ) -> CreateJobResponse: ...
    @abstractmethod
    def UpdateJob(
        self,
        request: UpdateJobRequest,
        context: ServicerContext,
    ) -> UpdateJobResponse: ...
    @abstractmethod
    def DeleteJob(
        self,
        request: DeleteJobRequest,
        context: ServicerContext,
    ) -> DeleteJobResponse: ...
    @abstractmethod
    def GetJobByName(
        self,
        request: GetJobByNameRequest,
        context: ServicerContext,
    ) -> GetJobByNameResponse: ...

def add_TopicPushServiceServicer_to_server(servicer: TopicPushServiceServicer, server: Server) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/server/v1/bootstrap.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.server.v1 import environment_pb2 as chalk_dot_server_dot_v1_dot_environment__pb2
from chalk._gen.chalk.server.v1 import team_pb2 as chalk_dot_server_dot_v1_dot_team__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1f\x63halk/server/v1/bootstrap.proto\x12\x0f\x63halk.server.v1\x1a!chalk/server/v1/environment.proto\x1a\x1a\x63halk/server/v1/team.proto"\xbe\x01\n!BootstrapExtraSettingsEnvironment\x12\\\n\x08settings\x18\x01 \x03(\x0b\x32@.chalk.server.v1.BootstrapExtraSettingsEnvironment.SettingsEntryR\x08settings\x1a;\n\rSettingsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x08R\x05value:\x02\x38\x01"\xf4\x02\n\x16\x42ootstrapExtraSettings\x12K\n\x06global\x18\x01 \x03(\x0b\x32\x33.chalk.server.v1.BootstrapExtraSettings.GlobalEntryR\x06global\x12]\n\x0c\x65nvironments\x18\x02 \x03(\x0b\x32\x39.chalk.server.v1.BootstrapExtraSettings.EnvironmentsEntryR\x0c\x65nvironments\x1a\x39\n\x0bGlobalEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x08R\x05value:\x02\x38\x01\x1as\n\x11\x45nvironmentsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12H\n\x05value\x18\x02 \x01(\x0b\x32\x32.chalk.server.v1.BootstrapExtraSettingsEnvironmentR\x05value:\x02\x38\x01"\xa8\x03\n\x16ParsedBootstrapConfigs\x12+\n\x05teams\x18\x01 \x03(\x0b\x32\x15.chalk.server.v1.TeamR\x05teams\x12\x34\n\x08projects\x18\x02 \x03(\x0b\x32\x18.chalk.server.v1.ProjectR\x08projects\x12@\n\x0c\x65nvironments\x18\x03 \x03(\x0b\x32\x1c.chalk.server.v1.EnvironmentR\x0c\x65nvironments\x12>\n\x0cteam_invites\x18\x04 \x03(\x0b\x32\x1b.chalk.server.v1.TeamInviteR\x0bteamInvites\x12N\n\x0e\x65xtra_settings\x18\x05 \x01(\x0b\x32\'.chalk.server.v1.BootstrapExtraSettingsR\rextraSettings\x12<\n\x18global_pinned_base_image\x18\x06 \x01(\tH\x00R\x15globalPinnedBaseImage\x88\x01\x01\x42\x1b\n\x19_global_pinned_base_imageB\x97\x01\n\x13\x63om.chalk.server.v1B\x0e\x42ootstrapProtoP\x01Z\x12server/v1;serverv1\xa2\x02\x03\x43SX\xaa\x02\x0f\x43halk.Server.V1\xca\x02\x0f\x43halk\\Server\\V1\xe2\x02\x1b\x43halk\\Server\\V1\\GPBMetadata\xea\x02\x11\x43halk::Server::V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.server.v1.bootstrap_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\023com.chalk.server.v1B\016BootstrapProtoP\001Z\022server/v1;serverv1\242\002\003CSX\252\002\017Chalk.Server.V1\312\002\017Chalk\\Server\\V1\342\002\033Chalk\\Server\\V1\\GPBMetadata\352\002\021Chalk::Server::V1"
    _globals["_BOOTSTRAPEXTRASETTINGSENVIRONMENT_SETTINGSENTRY"]._options = None
    _globals["_BOOTSTRAPEXTRASETTINGSENVIRONMENT_SETTINGSENTRY"]._serialized_options = b"8\001"
    _globals["_BOOTSTRAPEXTRASETTINGS_GLOBALENTRY"]._options = None
    _globals["_BOOTSTRAPEXTRASETTINGS_GLOBALENTRY"]._serialized_options = b"8\001"
    _globals["_BOOTSTRAPEXTRASETTINGS_ENVIRONMENTSENTRY"]._options = None
    _globals["_BOOTSTRAPEXTRASETTINGS_ENVIRONMENTSENTRY"]._serialized_options = b"8\001"
    _globals["_BOOTSTRAPEXTRASETTINGSENVIRONMENT"]._serialized_start = 116
    _globals["_BOOTSTRAPEXTRASETTINGSENVIRONMENT"]._serialized_end = 306
    _globals["_BOOTSTRAPEXTRASETTINGSENVIRONMENT_SETTINGSENTRY"]._serialized_start = 247
    _globals["_BOOTSTRAPEXTRASETTINGSENVIRONMENT_SETTINGSENTRY"]._serialized_end = 306
    _globals["_BOOTSTRAPEXTRASETTINGS"]._serialized_start = 309
    _globals["_BOOTSTRAPEXTRASETTINGS"]._serialized_end = 681
    _globals["_BOOTSTRAPEXTRASETTINGS_GLOBALENTRY"]._serialized_start = 507
    _globals["_BOOTSTRAPEXTRASETTINGS_GLOBALENTRY"]._serialized_end = 564
    _globals["_BOOTSTRAPEXTRASETTINGS_ENVIRONMENTSENTRY"]._serialized_start = 566
    _globals["_BOOTSTRAPEXTRASETTINGS_ENVIRONMENTSENTRY"]._serialized_end = 681
    _globals["_PARSEDBOOTSTRAPCONFIGS"]._serialized_start = 684
    _globals["_PARSEDBOOTSTRAPCONFIGS"]._serialized_end = 1108
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import collections.abc
import functools
import random
import weakref
from itertools import zip_longest
from typing import Any, Generator, Iterable, Iterator, Literal, Mapping, Sequence, TypeVar, cast

import pyarrow as pa
import pyarrow.compute
from typing_extensions import Self, assert_never

from chalk.utils.collections import FrozenOrderedSet
from chalk.utils.df_utils import pa_cast, pa_table_to_pl_df, to_pylist
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception
from chalk.utils.weak_set_by_identity import WeakSetByIdentity

_LEFT_INDEX_COL = "__chalk__.ChalkRecordBatch.LEFT_INDEX_COL"
_RIGHT_INDEX_COL = "__chalk__.ChalkRecordBatch.RIGHT_INDEX_COL"

# We are missing good type stubs for pyarrow compute
pc = cast(Any, pyarrow.compute)

# For small tables, no need to construct a sequential index list for each join
# Instead, we'll use this preallocated list, and slice from it
# By using 32 bit integers, this list will take up 4 * (2**24) bytes ~= 67 MB
_PREALLOCATED_INDEX_ARRAY_LEN = 2**24
_PREALLOCATED_INDEX_ARRAY = pc.subtract_checked(
    pc.cumulative_sum_checked(pa.nulls(_PREALLOCATED_INDEX_ARRAY_LEN, pa.int32()).fill_null(1)),
    pa.scalar(1, pa.int32()),
)

_MIN_NUM_ROWS_FOR_POLARS_JOIN = 10_000
"""Minimum number of rows on either side before converting the table to/from polars to use a polars join. Found that polars
becomes faster around 10_000 rows on my mac"""

T = TypeVar("T")


_MATERIALIZED_ARRAYS: dict[int, list[Any]] = {}
"""A cache containing the result of calling `.to_list()` on a PyArrow array.

This cache is indexed by the id(array), and a weakref finalizer is used to clear memory from this array
when the underlying pyarrow array is freed. This design, unlike passing the materialized values through
the ChalkRecordBatch, allows for new tables to reuse the existing materialization of a .to_pylist() call
if another ChalkRecordBatch had to materialize the same underlying array.
"""

_logger = get_logger(__name__)


def _clear_from_materialized(id: int):
    # FIXME: This if guard shouldn't be necessary. Added for CHA-2917
    if id in _MATERIALIZED_ARRAYS:
        del _MATERIALIZED_ARRAYS[id]
    else:
        _logger.warning(f"Could not find id {id} in the materialized arrays cache")


_sequential_index_column_set: WeakSetByIdentity[pa.Array] = WeakSetByIdentity()


def make_sequential_index_col(n: int) -> pa.Array:
    """
    Returns a `pa.Array` with exactly `n` rows, having values 0, 1, 2, 3, ...
    If possible, uses a slice of the _PREALLOCATED_INDEX_ARRAY to reduce memory overhead.

    The type of the returned array is pa.int64().
    """

    if n <= _PREALLOCATED_INDEX_ARRAY_LEN:
        array = _PREALLOCATED_INDEX_ARRAY.slice(0, n).cast(pa.int64())
    else:
        # The desired array is larger than the preallocated sequential list,
        # so we need to construct a new one.
        array = pc.subtract_checked(
            pc.cumulative_sum_checked(pa.nulls(n, pa.int64()).fill_null(pa.scalar(1, type=pa.int64()))),
            pa.scalar(1, pa.int64()),
        )

    _sequential_index_column_set.add(array)
    return array


def is_definitely_sequential_index_column(array: pa.Array) -> bool:
    """
    This function can be used for optimization - if `True` is returned, then the array
    is a sequential array of `0, 1, 2, 3, 4, ...` up to its length, allowing for rapid
    comparison.
    """
    return array in _sequential_index_column_set


@functools.lru_cache(None)
def _randomize_join_result_order():
    """Whether to randomize the results of ChalkRecordBatch.join. Useful during testing
    to ensure that we do not inadvertanetly assume that a ChalkRecordBatch.join does not change row order"""
    return env_var_bool("CHALK_TABLE_RANDOMIZE_JOIN_RESULT_ORDER")


def _randomize_row_order(tbl: ChalkRecordBatch) -> ChalkRecordBatch:
    if len(tbl) <= 1:
        return tbl
    offset = random.randint(0, len(tbl) - 1)
    len_scalar = pa.scalar(len(tbl), pa.int64())
    idxs = pc.add_checked(make_sequential_index_col(len(tbl)), pa.scalar(offset, pa.int64()))
    idxs = pc.if_else(pc.greater_equal(idxs, len_scalar), pc.subtract_checked(idxs, len_scalar), idxs)
    tbl = tbl.take(idxs)
    return tbl


class ChalkRecordBatch:
    """A collection of same-length PyArrow arrays, similar to a pa.RecordBatch, but with an easier to use API
    and better performance
    """

    # Private variables
    _data: Mapping[str, pa.Array]  # pyright: ignore[reportUninitializedInstanceVariable]
    _column_names: tuple[str, ...]  # pyright: ignore[reportUninitializedInstanceVariable]
    _len: int  # pyright: ignore[reportUninitializedInstanceVariable]
    _schema: Mapping[str, pa.DataType]  # pyright: ignore[reportUninitializedInstanceVariable]

    def __init__(self):
        super().__init__()
        raise RuntimeError(
            "Do not construct a ChalkRecordBatch directly. Instead, please use one of the ChalkRecordBatch.from_XXX methods."
        )

    def __repr__(self):
        """
        Render contents as a well-formatted table in debugging contexts.
        """
        equivalent_df = pa_table_to_pl_df(self.to_table())
        return repr(equivalent_df)

    def __hash__(self):
        # This isn't perfect, but it's better than always returning the same thing
        return hash((self._column_names, self._len))

    def __eq__(self, other: object):
        if not isinstance(other, ChalkRecordBatch):
            return NotImplemented
        return self.equals(other, check_column_order=True)

    @classmethod
    def _internal_constructor(
        cls,
        data: Mapping[str, pa.Array],
        columns: tuple[str, ...],
        data_len: int,
        schema: Mapping[str, pa.DataType],
    ) -> Self:
        self = super().__new__(cls)
        self._data = data
        self._column_names = columns
        self._len = data_len
        self._schema = schema
        if __debug__:
            assert tuple(data.keys()) == columns, "The order of the keys in the data dict must match the column names"
            assert (
                tuple(schema.keys()) == columns
            ), "The order of the keys in the schema dict must match the column names"
            assert FrozenOrderedSet(columns) == FrozenOrderedSet(
                data.keys()
            ), "The columns list must contain all keys in ``data``"
            assert all(isinstance(x, pa.Array) for x in data.values()), "All values must be pa.Array"
            assert len(columns) == len(data), "The columns contains duplicate entries"
            assert schema == (
                actual_schema := {k: v.type for (k, v) in data.items()}
            ), f"Schema mismatch. Declared {schema}, but got {actual_schema}"
            if len(self._data) == 0:
                assert self._len == 0
            else:
                assert all(len(x) == self._len for x in self._data.values()), "All arrays must have the same length"
        return self

    __slots__ = ("_data", "_column_names", "_len", "_schema")

    def __len__(self) -> int:
        return self._len

    def __getitem__(self, name: str) -> pa.Array:
        try:
            return self._data[name]
        except KeyError:
            raise ValueError(f"Column '{name}' not in the table") from None

    def append_column(self, field: str, column: pa.Array | pa.ChunkedArray):
        """Create a new table with this additional column at the end of the table

        Parameters
        ----------
        field
            The column name to append. A column with this name cannot currently be in the table
        column
            The PyArrow array containing the data

        Returns
        -------
        A new ChalkRecordBatch
        """
        return self.append_columns({field: column})

    def append_columns(
        self,
        columns: Mapping[str, pa.Array | pa.ChunkedArray | pa.Scalar],
    ):
        """Create a new table by appending the ``columns`` to the table.

        Column order is preserved, with new columns appearing at the end of the table.

        This API is more performant than creating a new ChalkRecordBatch from scratch or repeated ``append_column`` calls
        """
        if len(self.columns) > 0:
            expected_len = self._len
            expected_len_src = "the table"
        else:
            # If the table is 0x0, then use the length of the first column in columns for the length check
            # It's impossible to have an expression if the table is 0x0, since the expression must come from an existing column
            first_real_col = next(
                iter(x for x in columns.values() if isinstance(x, (pa.Array, pa.ChunkedArray))),
                None,
            )
            if first_real_col is None:
                raise ValueError("Cannot append only scalars to a 0x0 table")
            expected_len = len(first_real_col)
            expected_len_src = "the other appended columns"
        new_data = {**self._data}
        new_schema = {**self._schema}
        for col_name, column in columns.items():
            if not isinstance(col_name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column names must be strings")
            if isinstance(column, pa.Scalar):
                column = pa.nulls(expected_len, column.type).fill_null(column)
            if isinstance(column, pa.ChunkedArray):
                column = column.combine_chunks()
            if not isinstance(column, (pa.Array, pa.Scalar)):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column values must be PyArrow arrays, ChunkedArrays, or an expression")
            if len(column) != expected_len:
                raise ValueError(
                    f"Column {col_name} has a length of {len(column)}, which differs from the length of {expected_len_src} ({expected_len})"
                )
            if col_name in self._data:
                raise ValueError(
                    f"Column '{col_name}' is already in the table. Please use ``replace_columns`` to update existing columns in the table."
                )
            new_data[col_name] = column
            new_schema[col_name] = column.type
        return self._internal_constructor(
            new_data,
            tuple(new_data.keys()),
            expected_len or self._len,
            new_schema,
        )

    def column(self, idx_or_name: str | int) -> pa.Array:
        """Get a column from the table

        Parameters
        ----------
        idx_or_name
            The name or integer index of the column
        """
        if isinstance(idx_or_name, int):
            try:
                idx_or_name = self._column_names[idx_or_name]
            except IndexError:
                raise ValueError(
                    (
                        f"Index {idx_or_name} is not in the table. The table has {len(self._data)} columns, "
                        f"so the index must be between 0 and {len(self._data) - 1}"
                    )
                ) from None
        if not isinstance(idx_or_name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("idx_or_name must be the integer index or the column name string")
        try:
            return self._data[idx_or_name]
        except KeyError:
            raise ValueError(f"Column '{idx_or_name}' not found") from None

    @property
    def column_names(self) -> tuple[str, ...]:
        return self._column_names

    def contains_column_name(self, name: str) -> bool:
        return name in self._data

    @property
    def columns(self) -> tuple[pa.Array, ...]:
        # The order of data must be the same order of the columns
        return tuple(self._data.values())

    @property
    def data(self) -> Mapping[str, pa.Array]:
        """Get the columns as a mapping of {column_name: Array}"""
        return self._data

    def drop_columns(self, columns: str | Iterable[str]):
        if isinstance(columns, str):
            columns = (columns,)
        else:
            columns = tuple(columns)
        for column in columns:
            if not isinstance(column, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError(f"Each column name must be a string. Got {column}")
            if column not in self._data:
                raise ValueError(f"Column '{columns}' not in the table")
        new_datas = {k: v for (k, v) in self._data.items() if k not in columns}
        new_columns = tuple(x for x in self._column_names if x not in columns)
        new_schema = {k: v for (k, v) in self._schema.items() if k not in columns}
        new_len = 0 if len(new_datas) == 0 else self._len
        return self._internal_constructor(new_datas, new_columns, new_len, new_schema)

    def equals(self, other: ChalkRecordBatch, check_column_order: bool = True) -> bool:
        if self is other:
            return True
        if check_column_order:
            if other.column_names != self.column_names:
                return False
        else:
            if FrozenOrderedSet(self.column_names) != FrozenOrderedSet(other.column_names):
                return False
        if not isinstance(other, ChalkRecordBatch):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("ChalkRecordBatch.equals can only be used with another ChalkRecordBatch")
        if self._len != other._len:
            return False
        if self._schema != other._schema:
            return False
        ans = pa.nulls(self._len, pa.bool_()).fill_null(True)
        for col_name in self.column_names:
            ans = pc.and_(ans, pc.equal(self._data[col_name], other._data[col_name]))
        ret = pc.all(ans, min_count=0).as_py()
        assert isinstance(ret, bool)
        return ret

    def filter(self, mask: pa.Array | pa.ChunkedArray | pa.Scalar):
        if isinstance(mask, pa.Scalar):
            raise ValueError("A scalar mask doesn't make sense")
        if isinstance(mask, pa.ChunkedArray):
            mask = mask.combine_chunks()
        if not isinstance(mask, pa.Array):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"The mask must be an array, a chunked array, or a scalar; got {mask}")
        if mask.null_count > 0:
            raise ValueError("The mask cannot contain null values")
        if len(mask) != self._len:
            raise ValueError(
                f"The length of the mask ({len(mask)}) must be the same as the length of the table ({self._len})"
            )
        new_datas = {k: v.filter(mask) for (k, v) in self._data.items()}
        new_len = 0 if len(new_datas) == 0 else len(next(iter(new_datas.values())))
        return self._internal_constructor(new_datas, self.column_names, new_len, self._schema)

    @classmethod
    def from_arrays(
        cls,
        arrays: Iterable[pa.Array | pa.ChunkedArray],
        names: Iterable[str],
    ) -> Self:
        data: dict[str, pa.Array] = {}
        data_len = None
        names = tuple(names)
        if len(FrozenOrderedSet(names)) != len(names):
            raise ValueError(f"Duplicate name are not allowed: {names}")
        for name, arr in zip_longest(names, arrays, fillvalue=None):
            if name is None or arr is None:
                raise ValueError("The length of the arrays must match the length of the names")
            if not isinstance(name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("The column name must be a string")
            if isinstance(arr, pa.ChunkedArray):
                arr = arr.combine_chunks()
            if not isinstance(arr, pa.Array):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Each array must be an array or a chunked array")
            if pa.types.is_date32(arr.type):
                arr = arr.cast(pa.date64())
            if data_len is None:
                data_len = len(arr)
            elif len(arr) != data_len:
                raise ValueError(
                    f"Cannot construct a ChalkRecordBatch with different array lengths. {data_len=}, {len(arr)=}. Failed on column {name=}"
                )
            data[name] = arr
        if data_len is None:
            data_len = 0
        schema = {k: v.type for (k, v) in data.items()}
        return cls._internal_constructor(data, names, data_len, schema)

    @classmethod
    def from_pydict(
        cls,
        data: Mapping[str, Iterable[object] | pa.Array | pa.ChunkedArray],
        schema: Mapping[str, pa.DataType] | None = None,
    ) -> ChalkRecordBatch:
        new_data: dict[str, pa.Array] = {}
        if schema is None:
            new_schema: dict[str, pa.DataType] = {}
            data_len = None
            for k, v in data.items():
                if not isinstance(k, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                    raise TypeError(f"Column names must be strings; got {k}")
                if isinstance(v, pa.ChunkedArray):
                    v = v.combine_chunks()  # pyright: ignore[reportAttributeAccessIssue]
                if not isinstance(v, pa.Array):
                    raise TypeError("If no schema is provided, columns must be PyArrow arrays or chunked arrays")
                if data_len is None:
                    data_len = len(v)  # pyright: ignore[reportArgumentType]
                elif len(v) != data_len:  # pyright: ignore[reportArgumentType]
                    raise ValueError("Cannot construct a ChalkRecordBatch with different array lengths")
                new_data[k] = v
                new_schema[k] = v.type  # pyright: ignore[reportAttributeAccessIssue]
            if data_len is None:
                data_len = 0
            return cls._internal_constructor(new_data, tuple(data.keys()), data_len, new_schema)

        for k, v in schema.items():
            if not isinstance(k, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError(f"Schema keys must be strings. Got {k}")
            if not isinstance(v, pa.DataType):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError(f"Schema values must be pa.DataTypes. Got {v}")
        if FrozenOrderedSet(data.keys()) != FrozenOrderedSet(schema.keys()):
            raise ValueError("The `data` and `schema` must have the same keys")
        new_data: dict[str, pa.Array] = {}
        data_len = None
        for k in schema:
            v = data[k]
            if isinstance(v, (pa.Array, pa.ChunkedArray)):
                raise TypeError("When providing a schema, values cannot be existing pyarrow arrays")
            v = tuple(v)
            if data_len is None:
                data_len = len(v)
            elif len(v) != data_len:
                raise ValueError("Cannot construct a ChalkRecordBatch with different array lengths")

            v = pa.array(v, schema[k])
            if isinstance(v, pa.ChunkedArray):
                v = v.combine_chunks()
            new_data[k] = v
        if data_len is None:
            data_len = 0
        columns = tuple(schema.keys())  # Using the (implicit) ordering from the schema
        return cls._internal_constructor(new_data, columns, data_len, schema)

    @classmethod
    def from_table(cls, table: pa.Table | pa.RecordBatch) -> ChalkRecordBatch:
        if isinstance(table, cls):
            return table
        if not isinstance(table, (pa.Table, pa.RecordBatch)):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"The table must be an existing PyArrow table or RecordBatch; got {table}")
        if isinstance(table, pa.Table):
            column_names = tuple(table.column_names)
        else:
            column_names = tuple(table.schema.names)
        columns = tuple(table.columns)
        return cls._internal_constructor(
            {k: v if isinstance(v, pa.Array) else v.combine_chunks() for (k, v) in zip(column_names, columns)},
            column_names,
            0 if len(column_names) == 0 else len(table),
            {k: v.type for (k, v) in zip(column_names, columns)},
        )

    def get_total_buffer_size(self):
        return sum(x.get_total_buffer_size() for x in self._data.values())

    def itercolumns(self) -> Generator[pa.Array, None, None]:
        yield from self._data.values()

    @property
    def nbytes(self):
        return sum(x.nbytes for x in self._data.values())

    @property
    def num_columns(self):
        return len(self._data)

    @property
    def num_rows(self):
        return self._len

    def remove_column(self, name_or_idx: int | str):
        if isinstance(name_or_idx, int):
            if name_or_idx < 0:
                raise ValueError("Index must be >= 0")
            if name_or_idx >= len(self._data):
                raise ValueError(
                    f"Invalid index -- table has {len(self._data)} columns. Cannot drop index {name_or_idx}."
                )
            name_or_idx = self._column_names[name_or_idx]
        else:
            if not isinstance(name_or_idx, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError(f"Argument must be a string or int; got {type(name_or_idx)}")
            if name_or_idx not in self._data:
                raise ValueError(f"Column '{name_or_idx}' is not a valid column name")
        return self.drop_columns((name_or_idx,))

    def rename_columns(self, names: Iterable[str] | Mapping[str, str]):
        if isinstance(names, str):
            raise TypeError("names must be a mapping or an iterable of strings, not a singular string")

        if isinstance(names, collections.abc.Mapping):
            # This "cast" tells pyright to remember the generic arguments,
            # so that we can still perform value lookups.
            names_map: Mapping[str, str] = names
            for name in names:
                if name not in self._data:
                    raise ValueError(f"The table does not contain a column called '{name}'")
            names = tuple(names_map.get(k, k) for k in self._column_names)
        else:
            names = tuple(names)
            if len(names) != len(self._data):
                raise ValueError(
                    f"The new names (of length {len(names)}) do not match the number of columns in the table ({len(self._data)})"
                )
        # Since the data and schema are ordered dicts, we can simply iterate over them
        new_data = dict(zip(names, self._data.values()))
        new_schema = dict(zip(names, self._schema.values()))
        if len(new_data) != len(names):
            raise ValueError(f"New names contain duplicates: {names}")
        return self._internal_constructor(new_data, tuple(names), self._len, new_schema)

    @property
    def schema_dict(self) -> Mapping[str, pa.DataType]:
        return self._schema

    def set_column(self, i: int, field: str, column: pa.Array | pa.ChunkedArray):
        """Set the column at position ``i`` to have name ``field`` and value ``column``

        Creates a new table -- doesn't modify the table in-place
        """
        if not isinstance(i, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("The index must be an int")
        if not isinstance(field, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("The field must be a string")
        if not isinstance(column, (pa.Array, pa.ChunkedArray)):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("The replacement column must be a pyarrow array")
        if isinstance(column, pa.ChunkedArray):
            column = column.combine_chunks()
        if i < 0:
            raise ValueError("Negative indexes are not supported")
        if i >= len(self._data):
            raise ValueError(f"Index {i} is greater than the largest column index {len(self._column_names) - 1}")
        if len(column) != self._len:
            raise ValueError(
                f"The length of the new column ({len(column)}) differs from the existing table length ({self._len})"
            )
        new_data = {
            (field if i == j else c): (column if i == j else self._data[self._column_names[j]])
            for (j, c) in enumerate(self.column_names)
        }
        new_columns = tuple(field if i == j else c for (j, c) in enumerate(self._column_names))
        if len(FrozenOrderedSet(new_columns)) != len(new_columns):
            raise ValueError(f"Columns would contain a duplicate entry after this set operation: {new_columns}")
        new_schema = {
            (field if i == j else c): (column.type if i == j else self._schema[self._column_names[j]])
            for (j, c) in enumerate(self.column_names)
        }
        return self._internal_constructor(new_data, new_columns, self._len, new_schema)

    @property
    def shape(self) -> tuple[int, int]:
        return (self._len, len(self._data))

    def slice(self, offset: int, length: int | None = None) -> ChalkRecordBatch:
        """Slice the dataframe starting at ``offset`` for up to ``length`` rows

        If ``length`` is None, then all remaining rows (starting at the offset) are returned
        If ``length`` is greater than the number of remaining rows, it has the same effect as being called with ``None``
        It is not permitted for ``length`` to be negative.
        """
        if not isinstance(offset, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("The offset must be an int")
        if offset < 0:
            raise ValueError("Negative offsets are not supported")
        if offset > self._len:
            raise ValueError(f"Offset {offset} is greater than the length of the table {self._len}")
        if length is None:
            length = self._len - offset
        if not isinstance(length, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("The length must be an integer or None")
        if length < 0:
            raise ValueError("Negative lengths are not supported")
        if length > self._len - offset:
            length = self._len - offset
        if offset == 0 and length >= self._len:
            # Not actually slicing!
            return self
        return self._internal_constructor(
            {k: v.slice(offset, length) for (k, v) in self._data.items()},
            self._column_names,
            length,
            self._schema,
        )

    def take(self, indices: pa.Array | pa.ChunkedArray):
        """Create a new table with the specified indices"""
        if isinstance(indices, pa.ChunkedArray):
            indices = indices.combine_chunks()
        if not isinstance(indices, pa.Array):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("The indices must be an Array or a ChunkedArray")
        if indices.null_count > 0:
            raise ValueError("The indices for ``take`` cannot have any null values")
        new_data = {k: v.take(indices) for (k, v) in self._data.items()}
        return self._internal_constructor(new_data, self._column_names, len(indices), self._schema)

    def to_table(self) -> pa.Table:
        return pa.Table.from_pydict(self._data)

    def to_record_batch(self) -> pa.RecordBatch:
        return pa.RecordBatch.from_pydict(self._data)

    def to_pydict(self) -> Mapping[str, list[object]]:
        ans: dict[str, list[object]] = {}
        for k, v in self._data.items():
            materialized = _MATERIALIZED_ARRAYS.get(id(v))
            if materialized is None:
                try:
                    materialized = to_pylist(v)
                except Exception as e:
                    raise ValueError(f"while processing {k}\n{e}")
                _MATERIALIZED_ARRAYS[id(v)] = materialized
                weakref.finalize(v, _clear_from_materialized, id(v))
            ans[k] = materialized
        return ans

    def to_string(self, show_metadata: bool = False, preview_cols: int = 0) -> str:
        ans: list[str] = [f"ChalkRecordBatch (length={self._len})"]
        if show_metadata:
            for k, v in self._schema.items():
                ans.append(f"{k}: {str(v)}")
        ans.append("---")
        if preview_cols:
            for k, v in self._data.items():
                col_formatted = v.to_string(window=(preview_cols - 1) // 2 + 1, skip_new_lines=True)
                ans.append(f"{k}: {col_formatted}")
        return "\n".join(ans)

    def hstack(self, other: ChalkRecordBatch, right_suffix: str | None = None):
        """Create a new table by horizontally concatenating another table onto this one. Both tables must have the same length, unless
        if one of the tables is zero-by-zero, then the other table will be returned.

        Parameters
        ----------
        other
            The other table to combine
        right_suffix
            If there are any overlapping columns, a suffix to append to the names of the columns in the right table
            If not specified (or there are overlapping columns after applying this suffix), then a ValueError will be raised
        """
        if not isinstance(other, ChalkRecordBatch):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError("The other table must be a ChalkRecordBatch")
        if len(other.column_names) == 0:
            return self
        if len(self.column_names) == 0:
            return other
        if other._len != self._len:
            raise ValueError(f"The tables have different lengths. Left: {self._len}, Right: {other._len}")

        new_data = {**self._data}  # shallow copy
        new_schema = {**self._schema}  # shallow copy
        for k, v in other._data.items():
            new_col_name = k
            if k in new_data and right_suffix is not None:
                new_col_name = k + right_suffix
                if new_col_name in other._data:
                    raise ValueError(
                        f"Column '{k}' was renamed to '{new_col_name}', which already exists in the right table. Please try changing the ``right_suffix``"
                    )
            if new_col_name in new_data:
                raise ValueError(
                    f"Column '{k}' appears in both the original and hstacked table. Please try specifying the ``right_suffix``"
                )
            new_data[new_col_name] = v
            new_schema[new_col_name] = other._schema[k]
        new_columns = tuple(new_data.keys())
        return self._internal_constructor(new_data, new_columns, self._len, new_schema)

    def union_all(self, others: Sequence[ChalkRecordBatch]):
        if not all([x.schema_dict == self.schema_dict for x in others]):
            raise ValueError("All tables must have the same schema to union all")

        data: dict[str, pa.Array] = {}
        for k, v in self._data.items():
            other_v = [x._data[k] for x in others]
            data[k] = pa.concat_arrays([v] + other_v)

        return self._internal_constructor(
            data,
            self._column_names,
            len(self) + sum(len(x) for x in others),
            self._schema,
        )

    def replace_columns(
        self,
        columns: Mapping[str, pa.Array | pa.ChunkedArray | pa.Scalar],
    ):
        """Create a new table by replacing the ``columns``.

        Column order is preserved.

        This API is more performant than creating a new ChalkRecordBatch from scratch or repeated ``set_column` calls.
        """
        if len(columns) == 0:
            return self
        new_data = {**self._data}
        new_schema = {**self._schema}
        for col_name, column in columns.items():
            if not isinstance(col_name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column names must be strings")
            if isinstance(column, pa.Scalar):
                column = pa.nulls(self._len, column.type).fill_null(column)
            if isinstance(column, pa.ChunkedArray):
                column = column.combine_chunks()
            if not isinstance(column, pa.Array):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column values must be PyArrow arrays, ChunkedArrays, Scalars, or Expressions")
            if len(column) != self._len:
                raise ValueError(
                    f"Column '{col_name}' has a length of {len(column)}, which differs from the length of the table ({self._len})"
                )
            if col_name not in new_data:
                raise ValueError(
                    f"Column '{col_name}' is not already in the table. Please use ``append_columns`` to add new columns to the table"
                )
            new_data[col_name] = column
            new_schema[col_name] = column.type
        return self._internal_constructor(new_data, self._column_names, self._len, new_schema)

    def project(
        self,
        columns: Mapping[str, pa.Array | pa.ChunkedArray | pa.Scalar],
    ):
        if len(columns) == 0:
            return ChalkRecordBatch.from_pydict({}, {})
        new_data: dict[str, pa.Array] = {}
        new_schema: dict[str, pa.DataType] = {}
        if len(self.columns) > 0:
            expected_len = self._len
            expected_len_src = "the table"
        else:
            # If the table is 0x0, then use the length of the first column in columns for the length check
            # It's impossible to have an expression if the table is 0x0, since the expression must come from an existing column
            first_real_col = next(
                iter(x for x in columns.values() if isinstance(x, (pa.Array, pa.ChunkedArray))),
                None,
            )
            if first_real_col is None:
                raise ValueError("Cannot append only scalars to a 0x0 table")
            expected_len = len(first_real_col)
            expected_len_src = "the other appended columns"
        for col_name, column in columns.items():
            if not isinstance(col_name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column names must be strings")
            if isinstance(column, pa.Scalar):
                column = pa.nulls(expected_len, column.type).fill_null(column)
            if isinstance(column, pa.ChunkedArray):
                column = column.combine_chunks()
            if not isinstance(column, pa.Array):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column values must be PyArrow arrays or ChunkedArrays")
            if len(column) != expected_len:
                raise ValueError(
                    f"Column {col_name} has a length of {len(column)}, which differs from the length of {expected_len_src} ({expected_len})"
                )
            new_data[col_name] = column
            new_schema[col_name] = column.type
        return self._internal_constructor(new_data, tuple(new_data.keys()), expected_len, new_schema)

    def with_columns(
        self,
        columns: Mapping[str, pa.Array | pa.ChunkedArray | pa.Scalar],
    ):
        """Create a new table by appending (or replacing, if the columns currently exist), the ``columns`` to the table.
        It is similar to ``append_columns`` and ``replace_columns``, but does not validate whether or not the column name already appears in the table

        Column order is preserved, with new columns appearing at the end of the table.

        This API is more performant than creating a new ChalkRecordBatch from scratch or repeated ``append_column`` calls
        """
        if len(columns) == 0:
            return self
        new_data = {**self._data}
        new_schema = {**self._schema}
        if len(self.columns) > 0:
            expected_len = self._len
            expected_len_src = "the table"
        else:
            # If the table is 0x0, then use the length of the first column in columns for the length check
            # It's impossible to have an expression if the table is 0x0, since the expression must come from an existing column
            first_real_col = next(
                iter(x for x in columns.values() if isinstance(x, (pa.Array, pa.ChunkedArray))),
                None,
            )
            if first_real_col is None:
                raise ValueError("Cannot append only scalars to a 0x0 table")
            expected_len = len(first_real_col)
            expected_len_src = "the other appended columns"
        for col_name, column in columns.items():
            if not isinstance(col_name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column names must be strings")
            if isinstance(column, pa.Scalar):
                column = pa.nulls(expected_len, column.type).fill_null(column)
            if isinstance(column, pa.ChunkedArray):
                column = column.combine_chunks()
            if not isinstance(column, pa.Array):  # pyright: ignore[reportUnnecessaryIsInstance]
                raise TypeError("Column values must be PyArrow arrays or ChunkedArrays")
            if len(column) != expected_len:
                raise ValueError(
                    f"Column {col_name} has a length of {len(column)}, which differs from the length of {expected_len_src} ({expected_len})"
                )
            new_data[col_name] = column
            new_schema[col_name] = column.type
        return self._internal_constructor(new_data, tuple(new_data.keys()), expected_len, new_schema)

    def iter_pyrows(self, chunk_size: int = 1000) -> Iterator[dict[str, Any]]:
        """Iterate each row in the table as a python dictionary

        Parameters
        ----------
        chunk_size
            The number of rows to collect into python objects in one chunk. This parameter has no effect on correctness
        """
        if chunk_size < 1:
            raise ValueError("The chunk size must be at least 1")
        start = 0
        while start < self._len:
            slc = self.slice(start, chunk_size)
            start += len(slc)
            slc_dict = slc.to_pydict()
            for i in range(len(slc)):
                yield {k: v[i] for (k, v) in slc_dict.items()}

    def join(
        self,
        right_table: ChalkRecordBatch,
        keys: str | Iterable[str],
        join_type: Literal[
            "inner",
            "left outer",
            "merge",
        ],
        right_keys: str | Iterable[str] | None = None,
        right_suffix: str | None = None,
        use_threads: bool = True,
        unmatched_metadata: int | None = None,
        use_polars: bool | None = None,
    ) -> ChalkRecordBatch:
        """Create a new table by joining ``right_table`` against the current table.

        Unlike a PyArrow ``Table.join``, ``ChalkRecordBatch.join`` supports all data types, including list and struct columns
        In addition, our "inner" join isn't broken (see https://github.com/apache/arrow/issues/37729)
        NOTE: The returned column order is not guaranteed to be in any particular order

        Parameters
        ----------
        right_table
            The table to join against
        keys
            The columns in the left table to use as join keys
        join_type
            The type of join to perform
        right_keys
            The columns in the right table to use as the join keys. If not specified, the same keys as the keys will be used
        right_suffix
            A suffix to append to any non-join columns in the right table that are also in the left table. If not specified,
            or duplicate columns still exist after applying this suffix, a ValueError will be raised
        use_threads
            Whether to use threads for the join
        use_polars
            Whether to use polars for the join. If None, then heuristics will be used to determine if polars is faster
        unmatched_metadata
            In outer joins, what values to fill into the metadata for unmatched entries
        """
        # Pyarrow does not support joining on tables containing structs or lists in the non-join columns
        # To get around this limitation, we'll construct a pyarrow table for the left join keys and an index column.
        # We'll join this against a table containing the right join keys and an index column
        # Then, we'll hstack on the other non-join columns that match the indices from the left and right
        if not isinstance(right_table, ChalkRecordBatch):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"The right_table must be a ChalkRecordBatch; got {right_table}")
        if isinstance(keys, str):
            # If only one `str` key is given, convert it into a singleton list.
            # Remember that `str` is also an `Iterable[str]` (over characters), which could otherwise lead to confusion.
            keys = (keys,)
        else:
            keys = tuple(keys)
        if len(FrozenOrderedSet(keys)) != len(keys):
            raise ValueError(f"The left keys contained duplicate keys: {keys}")

        if right_keys is None:
            # If no `right_keys` are given, use `keys` as a fallback.
            right_keys = keys
        elif isinstance(right_keys, str):
            # If only one `str` key is given, convert it into a singleton list.
            # Remember that `str` is also an `Iterable[str]` (over characters), which could otherwise lead to confusion.
            right_keys = (right_keys,)
        else:
            right_keys = tuple(right_keys)
        if len(FrozenOrderedSet(right_keys)) != len(right_keys):
            raise ValueError(f"The right keys contained duplicate keys: {right_keys}")
        right_keys_set = set(right_keys)  # Build a set to make set-membership checks faster.

        # In order to perform a join, after we compute the key tuples, they must have the
        # same cardinality. Otherwise, it would be impossible for the left/right values to match.
        if len(keys) != len(right_keys):
            raise ValueError(f"The left keys ({keys}) must have the same length of the right keys ({right_keys})")

        # Ensure that all of the left keys exist in the left (self) table.
        missing_left_keys = tuple(x for x in keys if x not in self._data)
        if len(missing_left_keys) > 0:
            raise ValueError(f"The following join keys are not in the lhs table: {missing_left_keys}")

        # Ensure that all of the right keys exist in the right table.
        missing_right_keys = tuple(x for x in right_keys if x not in right_table._data)
        if len(missing_right_keys) > 0:
            raise ValueError(f"The following join keys are not in the lhs table: {missing_right_keys}")
        if join_type == "merge":
            if len(keys) != 1:
                raise ValueError("Merge join requires exactly one join key")
            if __debug__:
                assert pc.all(
                    pc.equal(
                        self.column(keys[0]),
                        right_table.column(right_keys[0]),
                    ),
                    min_count=0,
                ).as_py(), "Both sides must have an identical join column when merge joining"
            if len(self) != len(right_table):
                # This check is much cheaper to perform, even when PYTHONDEBUG=0
                raise ValueError(
                    f"Merge join requires identical columns when joining. The left hand side had {len(self)} elements, while the right hand side had {len(right_table)} elements"
                )

            # Merge join is a special case of left outer. Should implement the hstack codepath
            join_type = "left outer"

        if (
            len(keys) == 1
            and is_definitely_sequential_index_column(self._data[keys[0]])
            and is_definitely_sequential_index_column(right_table._data[right_keys[0]])
            and (
                len(self) == len(right_table)
                or join_type == "inner"
                or (len(self) < len(right_table) and join_type == "left outer")
                or (
                    len(self) > len(right_table)
                    and join_type == "right outer"  # pyright: ignore[reportUnnecessaryComparison]
                )  # pyright: ignore[reportUnnecessaryComparison]
            )
        ):
            # This is a fast path for joining two ChalkTables by a sequential column (0, 1, 2, 3, ...).
            #
            # If we know that both join columns are the sequential integers, then a join is the same as
            # just sticking all of the columns into one table. If the tables are the same length,
            # regardless of the specific type of merge, since there are no duplicates or missing values.
            # This works if the join type is such that there would never be nulls added (i.e. doing a left join
            # when there are more rows in the right)
            #
            # This should be a lot faster, since we don't have to even _look_ at the data!
            left_table = self
            if len(left_table) > len(right_table):
                left_table = left_table.slice(0, len(right_table))
            if len(left_table) < len(right_table):
                right_table = right_table.slice(0, len(left_table))

            add_columns_from_right: dict[str, pa.Array] = {}
            for right_column_name in right_table.column_names:
                if right_column_name == right_keys[0]:
                    # The join column in the right table gets skipped, since definitionally
                    # its values are the same as what came from the left table.
                    continue

                # By default, the right column name is used unchanged in the output.
                # Only if there's a conflict will we add a suffix.
                new_column_name = right_column_name

                if right_column_name in self.data:
                    if right_suffix is None:
                        raise ValueError(
                            f"Column {right_column_name} appears in both sides of the table. Please set the `right_suffix` to disambiguate."
                        )
                    new_column_name += right_suffix

                if new_column_name in self.data or new_column_name in add_columns_from_right:
                    raise ValueError(
                        f"Column '{right_column_name}' was renamed to '{new_column_name}', which already exists in the right table. Please try changing the ``right_suffix``"
                    )

                add_columns_from_right[new_column_name] = right_table._data[right_column_name]

            # Append all of the columns that we just added.
            res = left_table.append_columns(add_columns_from_right)
            if _randomize_join_result_order():
                res = _randomize_row_order(res)

            return res

        if use_polars is None:
            use_polars = len(self) > _MIN_NUM_ROWS_FOR_POLARS_JOIN or len(right_table) > _MIN_NUM_ROWS_FOR_POLARS_JOIN

        if (
            use_polars
            # Not allowing polars on list/struct types because I'm concerned about how it would handle nulls / empty lists and structs and recursion
            and not any(
                pa.types.is_struct(x)
                or pa.types.is_list(x)
                or pa.types.is_large_list(x)
                or pa.types.is_fixed_size_list(x)
                for x in self.schema_dict.values()
            )
            and not any(
                pa.types.is_struct(x)
                or pa.types.is_list(x)
                or pa.types.is_large_list(x)
                or pa.types.is_fixed_size_list(x)
                for x in right_table.schema_dict.values()
            )
        ):
            expected_schema = {**self.schema_dict}  # shallow copy
            for right_column_name, right_dtype in right_table.schema_dict.items():
                if right_column_name in right_keys_set:
                    # The join column in the right table gets skipped, since definitionally
                    # its values are the same as what came from the left table.
                    continue

                # By default, the right column name is used unchanged in the output.
                # Only if there's a conflict will we add a suffix.
                new_column_name = right_column_name

                if right_column_name in self.data:
                    if right_suffix is None:
                        raise ValueError(
                            f"Column {right_column_name} appears in both sides of the table. Please set the `right_suffix` to disambiguate."
                        )
                    new_column_name += right_suffix

                if new_column_name in expected_schema:
                    raise ValueError(
                        f"Column '{right_column_name}' was renamed to '{new_column_name}', which already exists in the right table. Please try changing the ``right_suffix``"
                    )
                expected_schema[new_column_name] = right_dtype

            left_df = pa_table_to_pl_df(self.to_table())
            right_df = pa_table_to_pl_df(right_table.to_table())
            if join_type == "inner":
                how = "inner"
            elif join_type == "left outer":
                how = "left"
            else:
                assert_never(join_type)
            df = left_df.join(
                right_df,
                how=how,
                left_on=keys,
                right_on=right_keys,
                suffix=right_suffix or "",
            )
            joined_table = df.to_arrow()
            res = ChalkRecordBatch.from_table(pa_cast(joined_table, pa.schema(expected_schema)))
            if _randomize_join_result_order():
                res = _randomize_row_order(res)
            return res

        # We need an "index" column to remember where each row in the join table came from.
        # If possible, we re-use a preallocated index array (0, 1, 2, 3, 4, ...), otherwise
        # we make a new one.
        left_idx_col = make_sequential_index_col(self._len)

        # The left_join_table has columns:
        # - one for each column in (left) "keys"
        # - plus the left_idx_col to remember where each row came from
        # The columns retain their original names, with `_LEFT_INDEX_COL` being a special name for the index col.
        left_join_table: pa.Table = pa.Table.from_arrays(
            [*(self._data[k] for k in keys), left_idx_col], [*keys, _LEFT_INDEX_COL]
        )
        del left_idx_col  # free memory; it will be kept alive in `left_join_table` only.

        # We need an "index" column to remember where each row in the join table came from.
        # If possible, we re-use a preallocated index array (0, 1, 2, 3, 4, ...), otherwise
        # we make a new one.
        right_idx_col = make_sequential_index_col(right_table._len)

        # The right_join_table has columns:
        # - one for each column in "right_keys"
        # - plus the right_idx_col to remember where each row came from
        # The columns retain their original names, with `_RIGHT_INDEX_COL` being a special name for the index col.
        right_join_table = pa.Table.from_arrays(
            [*(right_table._data[k] for k in right_keys), right_idx_col],
            [*right_keys, _RIGHT_INDEX_COL],
        )
        del right_idx_col  # free memory; it will be kept alive in `right_join_table` only.

        # Ask pyarrow to join the tables for us. We can recover the rest of the original rows later by using the special
        # `_LEFT_INDEX_COL` and `_RIGHT_INDEX_COL` columns.
        #
        # The resulting columns in `joined_table` are ONLY the (left) `keys` (not `right_keys`) and the two special
        # index columns `_LEFT_INDEX_COL` and `_RIGHT_INDEX_COL`.
        # NOTE: pyarrow inner joins are broken (see https://github.com/apache/arrow/issues/37729)
        # So, we will always do a "left outer" join, but if the join type is "inner", we will drop rows where the _RIGHT_INDEX_COL is null
        joined_table = left_join_table.join(
            right_table=right_join_table,
            keys=keys,
            right_keys=right_keys,
            join_type="left outer" if join_type == "inner" else join_type,
            # Since there aren't any other columns, there is no chance of a conflict.
            # Therefore, we don't need to set `right_suffix`.
            # The only other columns are `_LEFT_INDEX_COL` and `_RIGHT_INDEX_COL`, which
            # we don't want to be suffixes.
            use_threads=use_threads,
        )
        if join_type == "inner":
            joined_table = joined_table.filter(joined_table.column(_RIGHT_INDEX_COL).is_valid())
        del left_join_table, right_join_table  # free memory

        data_len = len(joined_table)

        # Convert the table back into a mapping from column-name to array.
        # Note: this only includes the (left) `keys` columns, since the `pyarrow.Table.join` operator
        # only includes the left side's key columns (which is where it gets the names).
        data: dict[str, pa.Array] = {
            col_name: col.combine_chunks() for (col_name, col) in zip(joined_table.column_names, joined_table.columns)
        }
        del joined_table  # free memory

        # We now assume that the `_LEFT_INDEX_COL` and `_RIGHT_INDEX_COL` columns have been scrambled
        # (order-wise) by the join, but they're now matched up exactly how we want to select from the
        # original tables. Remove them from `data` since we don't want them in the final output.
        left_indices = data.pop(_LEFT_INDEX_COL)
        right_indices = data.pop(_RIGHT_INDEX_COL)

        # The `schema` here will be for the output table, which includes the `key` columns, as well as
        # all of the original columns in the left and right tables, excluding anything from `right_keys`
        # in the right table.
        #
        # If there's a conflict between a column name in the left table and a column name in the right
        # table, `right_suffix` will be preprended to the right column. If it's `None` or this still
        # somehow results in a conflict, then we raise an exception.
        #
        # Initially, we just fill `schema` with the join key columns. All others are added subsequently.
        schema = {k: self._schema[k] for k in data}

        # Add in all of the columns in self (left table) that weren't used for the join.
        for col_name, col in self._data.items():
            if col_name in data:
                # It is a join key, so it is already present in the joined table.
                assert col_name in keys
            else:
                # Grab the corresponding rows from this column in the left table (self).
                taken = col.take(left_indices)
                if isinstance(taken, pa.ChunkedArray):
                    # It is unclear from the types whether this code is reachable or not.
                    taken = taken.combine_chunks()
                # Store this column in the output.
                data[col_name] = taken
                # Remember that this column is included in the output.
                schema[col_name] = self._schema[col_name]

        # Add in all of the columns in `right_table` that weren't already used for the join.
        # Note that `right_keys` are excluded below, so there is no chance for them to conflict.
        for col_name, col in right_table._data.items():
            if col_name in right_keys_set:
                # It is a join key, so it is already in `data`.
                continue

            # The resulting name will either be `col_name` if it does not conflict, or
            # `f"{right_suffix}{col_name}"` if it does. If that still conflicts or
            # `right_suffix` is `None`, then an error gets raised.
            # Thus we ensure that the resulting table has no duplicate columns and the
            # right table doesn't trample over any values from the left.
            final_col_name = col_name
            if col_name in data and right_suffix is not None:
                final_col_name = col_name + right_suffix
                if final_col_name in right_table._data:
                    raise ValueError(
                        f"Column '{col_name}' was renamed to '{final_col_name}', which already exists in the right table. Please try changing the ``right_suffix``"
                    )
            if final_col_name in data:
                raise ValueError(
                    f"Column {final_col_name} appears in both sides of the table. Please set the `right_suffix` to disambiguate."
                )
            # Grab the corresponding rows from this column in the left table (self).
            taken = col.take(right_indices)
            if isinstance(taken, pa.ChunkedArray):
                # It is unclear from the types whether this code is reachable or not.
                taken = taken.combine_chunks()
            # Store this column in the output.
            data[final_col_name] = taken
            # Remember that this column is included in the output.
            schema[final_col_name] = right_table.schema_dict[col_name]
        # Set the unmatched metadata
        if unmatched_metadata is not None:
            unmatched_metadata_scalar = pa.scalar(unmatched_metadata, pa.uint64())
            for col_name, col in tuple(data.items()):
                if col_name.startswith("__chalk__.__metadata__."):
                    col = col.fill_null(unmatched_metadata_scalar)
                    data[col_name] = col
        columns = tuple(data)
        if __debug__:
            assert len(data) == (
                expected_num_cols := len(self._data) + len(right_table._data) - len(keys)
            ), f"Resulting column mismatch. Expected {expected_num_cols} columns; got {len(data)}"
            for k in self._data:
                assert k in data, f"Missing column {k}"
            for k in right_table._data:
                assert (
                    k in data
                    or (k in right_keys and keys[right_keys.index(k)] in data)
                    or (right_suffix is not None and k + right_suffix in data)
                ), f"Missing column '{k}'"

        res = self._internal_constructor(data, columns, data_len, schema)
        if _randomize_join_result_order():
            res = _randomize_row_order(res)
        return res

    def __str__(self):
        return self.to_string(show_metadata=True, preview_cols=10)

    def pprint_string(self) -> str:
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[runtime]")

        return str(pl.from_arrow(self.to_table()))

    def with_row_count(self, name: str):
        """Append a row count column to the table. This column will be called ``name``, and the first row will have the ``start`` value"""
        return self.append_column(name, make_sequential_index_col(len(self)))

    def with_unique_id(self, name: str) -> Self:
        # When creating a new unique id index, we will randomize the start of it
        # This makes debugging easier to find when we should have used with_row_count()
        # Instead of with_unique_id()
        if len(self) == 0:
            # There's a bug with polars where roundtripping with a slice of length zero crashes
            return self.append_column(name, pa.nulls(0, pa.int64()))
        offset = random.randint(0, _PREALLOCATED_INDEX_ARRAY_LEN // 2)
        return self.append_column(name, make_sequential_index_col(len(self) + offset).slice(offset, len(self)))

    def sort_by(self, by: Iterable[tuple[str, Literal["ascending", "descending"]]]):
        by = list(by)
        if len(by) == 0:
            raise ValueError("Must have at least one column to sort by")
        return ChalkRecordBatch.from_table(self.to_table().sort_by(by))

    def distinct(
        self,
        on: Sequence[str] | None,
        order_by: Iterable[tuple[str, Literal["ascending", "descending"]]],
    ):
        order_by = list(order_by)
        table = self
        if len(order_by) > 0:
            table = self.sort_by(order_by)
        df = pa_table_to_pl_df(table.to_table())
        df = df.unique(on, keep="first")
        return ChalkRecordBatch.from_table(
            pa_cast(
                df.to_arrow(),
                pa.schema(self.schema_dict),
            )
        )

    def aggregate(self, by: Sequence[str], metrics: Sequence[tuple[str, str]]):
        return ChalkRecordBatch.from_table(self.to_table().group_by(list(by)).aggregate(list(metrics)))

    def select(self, cols: str | Iterable[str]) -> Self:
        if isinstance(cols, str):
            cols = (cols,)
        else:
            cols = tuple(cols)
        new_data: dict[str, pa.Array] = {}
        new_schema: dict[str, pa.DataType] = {}
        for c in cols:
            col = self._data.get(c)
            if col is None:
                raise ValueError(f"Column '{c}' does not exist")
            if c in new_data:
                raise ValueError(f"Column '{c}' cannot appear twice")
            new_data[c] = col
            new_schema[c] = self._schema[c]
        return self._internal_constructor(
            new_data,
            tuple(new_data.keys()),
            0 if len(new_data) == 0 else len(self),
            new_schema,
        )


def assert_table_equal(
    left: ChalkRecordBatch,
    right: ChalkRecordBatch,
    *,
    check_row_order: bool = True,
    check_column_order: bool = True,
):
    try:
        import polars.testing
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    polars.testing.assert_frame_equal(
        pa_table_to_pl_df(left.to_table()),
        pa_table_to_pl_df(right.to_table()),
        check_row_order=check_row_order,
        check_column_order=check_column_order,
    )



from __future__ import annotations

import contextlib
import functools
import queue
from datetime import date, datetime, time
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Iterator, Mapping, Optional, Sequence, cast
from uuid import uuid4

import google
import pyarrow as pa

from chalk.clogging import chalk_logger
from chalk.features import Feature
from chalk.integrations.named import create_integration_variable, load_integration_variable
from chalk.sql._internal.query_execution_parameters import QueryExecutionParameters
from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind, validate_dtypes_for_efficient_execution
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import google.cloud.bigquery
    import sqlalchemy
    from google.cloud.bigquery import ScalarQueryParameterType
    from sqlalchemy.engine import Connection
    from sqlalchemy.engine.url import URL
    from sqlalchemy.sql.ddl import CreateTable, DropTable


try:
    import sqlalchemy as sa
except ImportError:
    sa = None

if sa is None:
    _supported_sqlalchemy_types_for_pa_querying = ()
else:
    _supported_sqlalchemy_types_for_pa_querying = (
        sa.BigInteger,
        sa.Boolean,
        sa.BINARY,
        sa.BLOB,
        sa.LargeBinary,
        sa.Float,
        sa.Integer,
        sa.Time,
        sa.String,
        sa.Text,
        sa.VARBINARY,
        sa.DateTime,
        sa.Date,
        sa.SmallInteger,
        sa.BIGINT,
        sa.BOOLEAN,
        sa.CHAR,
        sa.DATETIME,
        sa.FLOAT,
        sa.INTEGER,
        sa.SMALLINT,
        sa.TEXT,
        sa.TIMESTAMP,
        sa.VARCHAR,
    )

_MAX_STATEMENT_LENGTH = 1024 * 1024
"""The actual maximum statement length that bigquery will accept. If the query is much longer than this, then we won't even
get a result from the server. This is likely because they refuse to process a request body this big!"""

_DEADLOCK_CIRCUMVENTING_TIMEOUT = 1
"""Used in a queue.put call to help circumvent a non-deterministic deadlock. See the function we're
using to monkeypatch the deadlocking BQ table download function for detailed explanation of the deadlock."""

_logger = get_logger(__name__)


@functools.lru_cache(maxsize=None)
def _determine_is_read_session_optional():
    # Used by monkeypatched BQ table download function.
    # If this variable fails to be resolved (remains as None)
    # then we will not monkeypatch the function.
    _is_read_session_optional = None
    try:
        import google.cloud.bigquery  # pyright: ignore[reportUnusedImport]
    except ModuleNotFoundError:
        # the user did not specify that they want to use the bigquery module
        pass
    else:
        try:
            from google.cloud.bigquery._versions_helpers import BQ_STORAGE_VERSIONS

            _is_read_session_optional = BQ_STORAGE_VERSIONS.is_read_session_optional
        except:
            try:
                # If customer's requirements.txt pins to some older version of bigquery
                from google.cloud.bigquery._helpers import (
                    BQ_STORAGE_VERSIONS,  # pyright: ignore[reportAttributeAccessIssue]
                )

                _is_read_session_optional = BQ_STORAGE_VERSIONS.is_read_session_optional
            except:
                # Failed to resolve the _is_read_session_optional variable.
                # We will evaluate the variable below and not monkeypatch
                # the BQ table download function.
                pass
    return _is_read_session_optional


def _download_table_bqstorage_stream_no_deadlock(
    download_state,  # pyright: ignore[reportMissingParameterType]
    bqstorage_client,  # pyright: ignore[reportMissingParameterType]
    session,  # pyright: ignore[reportMissingParameterType]
    stream,  # pyright: ignore[reportMissingParameterType]
    worker_queue,  # pyright: ignore[reportMissingParameterType]
    page_to_item,  # pyright: ignore[reportMissingParameterType]
):
    # Helps circumvent a non-deterministic deadlock. This deadlock could happen when:
    # 1. we are looping over the result of `to_arrow_iterable`
    # 2. an exception is raised in one of the loops
    # 3. we exit the loop, causing iterable consumption to stop, and queue popping
    #    to stop. At this point, the queue could be full. This means the worker
    #    thread would be blocked when calling the `Queue.put` method, and we never
    #    get to evaluate the download state and return.
    #
    # The following is the deadlocking portion of the original function, and is the
    # only portion that changed:
    #
    #    for page in rowstream.pages:
    #        if download_state.done:
    #            return
    #        item = page_to_item(page)
    #        worker_queue.put(item)
    #
    reader = bqstorage_client.read_rows(stream.name)

    # Avoid deprecation warnings for passing in unnecessary read session.
    # https://github.com/googleapis/python-bigquery-storage/issues/229
    if _determine_is_read_session_optional():
        rowstream = reader.rows()
    else:
        rowstream = reader.rows(session)

    for page in rowstream.pages:
        item = page_to_item(page)
        while True:
            if download_state.done:
                return
            try:
                worker_queue.put(item, timeout=_DEADLOCK_CIRCUMVENTING_TIMEOUT)
            except queue.Full:
                continue
            else:
                break


def monkeypatch_bq_table_download():
    try:
        import google.cloud.bigquery
    except ModuleNotFoundError:
        # the user did not specify that they want to use the bigquery module
        pass
    else:
        _is_read_session_optional = _determine_is_read_session_optional()
        # _is_read_session_optional being `None` means we failed to resolve the variable
        # due to import issues related to the BQ module. In that case, don't monkeypatch.
        _should_fallback = env_var_bool("CHALK_FALLBACK_TO_ORIGINAL_TABLE_DOWNLOAD_FUNC")
        if _should_fallback or _is_read_session_optional is None:
            _logger.warning(
                f"Not monkeypatching BQ table download function {_is_read_session_optional=} {_should_fallback=}"
            )
        else:
            try:
                import google.cloud.bigquery._pandas_helpers
            except ImportError:
                _logger.warning("Failed to import google.cloud.bigquery._pandas_helpers. Skipping monkey-patch.")
                pass
            else:
                _download_func_name = "_download_table_bqstorage_stream"
                if hasattr(google.cloud.bigquery._pandas_helpers, _download_func_name):
                    _logger.info("Monkeypatching BQ table download function")
                    setattr(
                        google.cloud.bigquery._pandas_helpers,
                        _download_func_name,
                        _download_table_bqstorage_stream_no_deadlock,
                    )
                else:
                    _logger.warning(
                        (
                            f"No {_download_func_name} function found in google.cloud.bigquery._pandas_helpers. "
                            "Skipping monkey-patch."
                        )
                    )


def _compile_parameter_to_bq_scalar_type(primitive: Any) -> ScalarQueryParameterType:
    # gated by import check in caller
    from google.cloud.bigquery import SqlParameterScalarTypes

    if isinstance(primitive, bool):
        return SqlParameterScalarTypes.BOOLEAN
    elif isinstance(primitive, int):
        return SqlParameterScalarTypes.INT64
    elif isinstance(primitive, float):
        return SqlParameterScalarTypes.FLOAT64
    elif isinstance(primitive, str):
        return SqlParameterScalarTypes.STRING
    elif isinstance(primitive, datetime):
        return SqlParameterScalarTypes.DATETIME
    elif isinstance(primitive, date):
        return SqlParameterScalarTypes.DATE
    elif isinstance(primitive, Decimal):
        return SqlParameterScalarTypes.DECIMAL
    elif isinstance(primitive, time):
        return SqlParameterScalarTypes.TIME
    elif isinstance(primitive, bytes):
        return SqlParameterScalarTypes.BYTES
    else:
        raise TypeError(f"Unsupported BigQuery parameter type '{type(primitive)}'")


_BQ_LOCATION_NAME = "BQ_LOCATION"
_BQ_DATASET_NAME = "BQ_DATASET"
_BQ_PROJECT_NAME = "BQ_PROJECT"
_BQ_CREDENTIALS_BASE64_NAME = "BQ_CREDENTIALS_BASE64"
_BQ_CREDENTIALS_PATH_NAME = "BQ_CREDENTIALS_PATH"


class BigQuerySourceImpl(BaseSQLSource):
    kind = SQLSourceKind.bigquery

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        project: Optional[str] = None,
        dataset: Optional[str] = None,
        location: Optional[str] = None,
        credentials_base64: Optional[str] = None,
        credentials_path: Optional[str] = None,
        engine_args: Optional[Dict[str, Any]] = None,
        integration_variable_override: Optional[Mapping[str, str]] = None,
    ):
        try:
            import sqlalchemy_bigquery
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[bigquery]")
        del sqlalchemy_bigquery  # unused
        monkeypatch_bq_table_download()
        if engine_args is None:
            engine_args = {}
        engine_args.setdefault("pool_size", 20)
        engine_args.setdefault("max_overflow", 60)
        self.location = location or load_integration_variable(
            integration_name=name, name=_BQ_LOCATION_NAME, override=integration_variable_override
        )
        self.dataset = dataset or load_integration_variable(
            integration_name=name, name=_BQ_DATASET_NAME, override=integration_variable_override
        )
        self.project = project or load_integration_variable(
            integration_name=name, name=_BQ_PROJECT_NAME, override=integration_variable_override
        )
        self.credentials_base64 = credentials_base64 or load_integration_variable(
            integration_name=name, name=_BQ_CREDENTIALS_BASE64_NAME, override=integration_variable_override
        )
        self.credentials_path = credentials_path or load_integration_variable(
            integration_name=name, name=_BQ_CREDENTIALS_PATH_NAME, override=integration_variable_override
        )
        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args={})

    @functools.cached_property
    def bigquery_read_client(self):
        import google.cloud.bigquery_storage

        with self._get_bq_client() as client:
            return google.cloud.bigquery_storage.BigQueryReadClient(
                credentials=client._credentials,  # pyright: ignore
            )

    def get_sqlglot_dialect(self) -> str | None:
        return "bigquery"

    def supports_inefficient_fallback(self) -> bool:
        return env_var_bool("CHALK_BIGQUERY_SUPPORTS_INEFFICIENT_FALLBACK", default=True)

    def compile_query(
        self,
        finalized_query: FinalizedChalkQuery,
        paramstyle: Optional[str] = None,
        use_sqlglot: bool = False,
    ) -> tuple[str, Sequence[Any], Dict[str, Any]]:
        if use_sqlglot:
            compiled_query = self._get_compiled_query(finalized_query, paramstyle)
            query_string = compiled_query.string

            import sqlglot.expressions
            from sqlglot import parse_one

            ast = parse_one(query_string, read=self.get_sqlglot_dialect())
            for placeholder in list(ast.find_all(sqlglot.expressions.Placeholder)):
                if isinstance(placeholder.this, str) and placeholder.this in compiled_query.params:
                    # Convert placeholders to use @ syntax
                    # https://cloud.google.com/bigquery/docs/parameterized-queries
                    placeholder.replace(sqlglot.expressions.var("@" + placeholder.this))
            updated_query_string = ast.sql(dialect="bigquery")

            return updated_query_string, compiled_query.positiontup, compiled_query.params
        else:
            if paramstyle is not None:
                raise ValueError("Bigquery does not support custom param styles")
            import google.cloud.bigquery
            import google.cloud.bigquery.dbapi
            import google.cloud.bigquery.dbapi._helpers
            import google.cloud.bigquery.dbapi.cursor
            from sqlalchemy_bigquery.base import BigQueryCompiler

            dialect = self.get_engine().dialect
            Compiler = cast("type[BigQueryCompiler]", self.get_engine().dialect.statement_compiler)

            compiled_stmt = Compiler(dialect, finalized_query.query)
            operation = compiled_stmt.string
            parameters = finalized_query.params if finalized_query.params else compiled_stmt.params
            (
                formatted_operation,
                parameter_types,
            ) = google.cloud.bigquery.dbapi.cursor._format_operation(  # pyright: ignore[reportPrivateUsage]
                operation, parameters
            )
            if len(formatted_operation) > _MAX_STATEMENT_LENGTH:
                raise ValueError("Query string is too large. Max supported size is 1MB.")
            # The DB-API uses the pyformat formatting, since the way BigQuery does
            # query parameters was not one of the standard options. Convert both
            # the query and the parameters to the format expected by the client
            # libraries.
            query_parameters = google.cloud.bigquery.dbapi._helpers.to_query_parameters(parameters, parameter_types)
            return formatted_operation, query_parameters, {}

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        query = {
            k: v
            for k, v in {
                "location": self.location,
                "credentials_base64": self.credentials_base64,
                "credentials_path": self.credentials_path,
            }.items()
            if v is not None
        }
        return URL.create(drivername="bigquery", host=self.project, database=self.dataset, query=query)

    @contextlib.contextmanager
    def _get_bq_client(self):
        # gated already
        import google.cloud.bigquery
        import google.cloud.bigquery.dbapi

        with self.get_engine().connect() as conn:
            dbapi = conn.connection.dbapi_connection
            assert isinstance(dbapi, google.cloud.bigquery.dbapi.Connection)
            client = dbapi._client  # pyright: ignore[reportPrivateUsage]
            assert isinstance(client, google.cloud.bigquery.Client)
            try:
                yield client
            finally:
                client.close()

    def _postprocess_table(
        self,
        features: Mapping[str, Feature],
        table: pa.RecordBatch,
    ):
        columns: list[pa.Array] = []
        column_names: list[str] = []

        for col_name, feature in features.items():
            column = table.column(col_name)
            expected_type = feature.converter.pyarrow_dtype
            if column.type != expected_type:
                column = column.cast(expected_type)
            if isinstance(column, pa.ChunkedArray):
                column = column.combine_chunks()
            columns.append(column)
            column_names.append(feature.root_fqn)
        return pa.RecordBatch.from_arrays(arrays=columns, names=column_names)

    @contextlib.contextmanager
    def _create_bigquery_temp_table(
        self,
        create_temp_table: CreateTable,
        temp_table: sqlalchemy.Table,
        drop_temp_table: DropTable,
        connection: google.cloud.bigquery.Client,
        temp_value: pa.Table,
        session_id: str | None,
    ):
        try:
            import google.cloud.bigquery
            import google.cloud.bigquery._pandas_helpers
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[bigquery]")

        create_table_sql = create_temp_table.compile(dialect=self.get_sqlalchemy_dialect()).string
        create_table_sql = create_table_sql.replace("TEMPORARY", "", 1)
        chalk_logger.info(f"Creating temporary table {temp_table.name} in BigQuery {session_id=}: {create_table_sql}")
        if session_id is not None:
            connection_properties = [google.cloud.bigquery.ConnectionProperty("session_id", session_id)]
        else:
            connection_properties = []
        job_config = google.cloud.bigquery.QueryJobConfig(
            priority="INTERACTIVE", connection_properties=connection_properties
        )
        connection.query(
            create_table_sql,
            job_config=job_config,
        ).result()
        try:
            connection.load_table_from_dataframe(
                temp_value.to_pandas(),
                f"{self.project}.{self.dataset}.{temp_table.name}",
                job_config=google.cloud.bigquery.LoadJobConfig(connection_properties=connection_properties),
            ).result()
            yield
        finally:
            drop_table_sql = drop_temp_table.compile(dialect=self.get_sqlalchemy_dialect()).string
            chalk_logger.info(f"Dropping temporary table {temp_table.name} in BigQuery: {drop_table_sql}")
            try:
                connection.query(
                    drop_table_sql,
                    job_config=job_config,
                ).result()
            except Exception as e:
                chalk_logger.warning(f"Failed to drop temporary BigQuery table {temp_table.name}: {e}")

    @contextlib.contextmanager
    def _bigquery_output_table(self, client: google.cloud.bigquery.Client) -> Iterator[str]:
        destination_table_name = f"temp_output_{str(uuid4()).replace('-', '_')}"

        destination = f"{client.project}.{self.dataset}.{destination_table_name}"

        try:
            yield destination
        finally:
            try:
                client.query(f"drop table {destination}").result()
            except Exception as e:
                chalk_logger.warning(f"Failed to drop temporary BigQuery table {destination}: {e}")

    def _execute_query_efficient(
        self,
        finalized_query: FinalizedChalkQuery,
        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
        connection: Optional[Connection],
        query_execution_parameters: QueryExecutionParameters,
    ) -> Iterable[pa.RecordBatch]:
        try:
            import google.cloud.bigquery
            import google.cloud.bigquery._pandas_helpers
            from sqlalchemy.sql import Select
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[bigquery]")

        if isinstance(finalized_query.query, Select):
            validate_dtypes_for_efficient_execution(finalized_query.query, _supported_sqlalchemy_types_for_pa_querying)

        client: google.cloud.bigquery.Client
        with self._get_bq_client() as client:
            query_job = client.query(
                "select 1;",
                job_config=google.cloud.bigquery.QueryJobConfig(priority="INTERACTIVE", create_session=True),
            )
            if query_job.session_info is not None:
                session_id = query_job.session_info.session_id
            else:
                session_id = None
            if session_id is None:
                _logger.warning("Failed to create a session, which is required for temp tables.")
            else:
                _logger.info(f"Created {session_id=}.")
            chalk_logger.info("Starting to execute BigQuery query")
            with contextlib.ExitStack() as exit_stack:
                for (
                    _,
                    temp_value,
                    create_temp_table,
                    temp_table,
                    drop_temp_table,
                ) in finalized_query.temp_tables.values():
                    exit_stack.enter_context(
                        self._create_bigquery_temp_table(
                            create_temp_table, temp_table, drop_temp_table, client, temp_value, session_id
                        )
                    )

                formatted_op, positional_params, named_params = self.compile_query(finalized_query)
                if named_params:
                    positional_params = [
                        # TODO: Consider type_ parameter more carefully.
                        google.cloud.bigquery.ScalarQueryParameter(
                            name=name, value=value, type_=_compile_parameter_to_bq_scalar_type(value)
                        )
                        for name, value in named_params.items()
                    ]
                if session_id is not None:
                    connection_properties = [google.cloud.bigquery.ConnectionProperty("session_id", session_id)]
                else:
                    connection_properties = []

                with self._bigquery_output_table(client) as destination:
                    job_config = google.cloud.bigquery.QueryJobConfig(
                        priority="INTERACTIVE",
                        query_parameters=positional_params or [],
                        connection_properties=connection_properties,
                        destination=destination,
                    )

                    chalk_logger.info(f"Executing BigQuery query: {formatted_op}")

                    res = client.query(formatted_op, job_config=job_config).result()
                    yielded = False

                    total_rows = 0
                    total_bytes = 0

                    for table in res.to_arrow_iterable(bqstorage_client=self.bigquery_read_client):
                        assert isinstance(table, pa.RecordBatch)
                        chalk_logger.info(f"Loaded table from Bigquery with {table.nbytes=}, {table.num_rows=}")
                        total_rows += table.num_rows
                        total_bytes += table.nbytes
                        features = columns_to_features(table.schema.names)
                        yield self._postprocess_table(features, table)
                        yielded = True
                    if not yielded and query_execution_parameters.yield_empty_batches:
                        # Copied from https://github.com/googleapis/python-bigquery/blob/89dfcb6469d22e78003a70371a0938a6856e033c/google/cloud/bigquery/table.py#L1954
                        arrow_schema = google.cloud.bigquery._pandas_helpers.bq_to_arrow_schema(
                            res._schema  # pyright: ignore[reportPrivateUsage]
                        )
                        if arrow_schema is not None:
                            features = columns_to_features(arrow_schema.names)
                            yield self._postprocess_table(
                                features,
                                pa.RecordBatch.from_pydict({k: [] for k in arrow_schema.names}, schema=arrow_schema),
                            )
                    chalk_logger.info(f"Loaded {total_rows=} rows and {total_bytes=} bytes from BigQuery")

    @classmethod
    def register_sqlalchemy_compiler_overrides(cls):
        try:
            from chalk.sql._internal.integrations.bigquery_compiler_overrides import register_bigquery_compiler_hooks
        except ModuleNotFoundError:
            raise missing_dependency_exception("chalkpy[bigquery]")

        register_bigquery_compiler_hooks()

    def _recreate_integration_variables(self) -> dict[str, str]:
        return {
            k: v
            for k, v in [
                create_integration_variable(_BQ_LOCATION_NAME, self.name, self.location),
                create_integration_variable(_BQ_DATASET_NAME, self.name, self.dataset),
                create_integration_variable(_BQ_PROJECT_NAME, self.name, self.project),
                create_integration_variable(_BQ_CREDENTIALS_BASE64_NAME, self.name, self.credentials_base64),
                create_integration_variable(_BQ_CREDENTIALS_PATH_NAME, self.name, self.credentials_path),
            ]
            if v is not None
        }



from chalk._gen.chalk.auth.v1 import permissions_pb2 as _permissions_pb2
from chalk._gen.chalk.common.v1 import feature_values_pb2 as _feature_values_pb2
from chalk._gen.chalk.common.v1 import query_log_pb2 as _query_log_pb2
from chalk._gen.chalk.common.v1 import query_values_pb2 as _query_values_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar

DESCRIPTOR: _descriptor.FileDescriptor

class OfflineStoreServiceDummy(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...



from __future__ import annotations

import asyncio
import os
import types
import warnings
from concurrent.futures import ThreadPoolExecutor
from datetime import timedelta
from functools import partial
from typing import Any, Mapping, Optional, TypeVar, Union
from uuid import UUID

import requests
import requests.adapters

from chalk.client import AsyncChalkClient
from chalk.client.client_impl import ChalkAPIClientImpl, OnlineQueryResponseImpl
from chalk.client.dataset import Dataset, DatasetImpl
from chalk.client.models import (
    BulkOnlineQueryResponse,
    ChalkError,
    FeatureStatisticsResponse,
    UploadFeaturesResponse,
    WhoAmIResponse,
)
from chalk.features.tag import BranchId, DeploymentId, EnvironmentId
from chalk.utils.log_with_context import get_logger
from chalk.utils.threading import ChalkThreadPoolExecutor

_logger = get_logger(__name__)

T = TypeVar("T")


class AsyncChalkClientImpl(AsyncChalkClient):
    __name__ = "AsyncChalkClient"
    __qualname__ = "chalk.client.AsyncChalkClient"

    def __repr__(self):
        return self._client.__repr__().replace("chalk.client.ChalkClient", "chalk.client.AsyncChalkClient")

    def __new__(cls, *args: Any, **kwargs: Any) -> AsyncChalkClientImpl:
        return object.__new__(AsyncChalkClientImpl)

    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        environment: Optional[EnvironmentId] = None,
        api_server: Optional[str] = None,
        query_server: Optional[str] = None,
        branch: Optional[BranchId] = None,
        preview_deployment_id: Optional[DeploymentId] = None,
        additional_headers: Optional[Mapping[str, str]] = None,
        default_job_timeout: Optional[Union[float, timedelta]] = None,
        default_request_timeout: Optional[Union[float, timedelta]] = None,
        executor: Optional[ThreadPoolExecutor] = None,
        pool_maxsize: Optional[int] = None,
        client: ChalkAPIClientImpl | None = None,
    ):
        self._executor = executor or ASYNC_CHALK_CLIENT_EXECUTOR
        if client is None:
            session = requests.Session()
            if pool_maxsize is not None:
                adapter = requests.adapters.HTTPAdapter(pool_maxsize=pool_maxsize)
                session.mount("http://", adapter)
                session.mount("https://", adapter)
            client = ChalkAPIClientImpl(
                client_id=client_id,
                client_secret=client_secret,
                environment=environment,
                api_server=api_server,
                query_server=query_server,
                branch=branch,
                preview_deployment_id=preview_deployment_id,
                session=session,
                additional_headers=additional_headers,
                default_request_timeout=default_request_timeout,
                default_job_timeout=default_job_timeout,
            )
        self._client = client

    async def __aenter__(self):
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: types.TracebackType | None,
    ):
        return

    def __enter__(self):
        warnings.warn(DeprecationWarning("Use `async with`, not `with`, with the AsyncChalkClient"))
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: types.TracebackType | None,
    ):
        return

    def _current_loop(self):
        return asyncio.get_running_loop()

    async def whoami(self) -> WhoAmIResponse:
        return await self._current_loop().run_in_executor(
            self._executor,
            self._client.whoami,
        )

    async def query(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> OnlineQueryResponseImpl:
        return await self._current_loop().run_in_executor(
            self._executor,
            partial(self._client.query, *args, **kwargs),
        )

    async def multi_query(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> BulkOnlineQueryResponse:
        return await self._current_loop().run_in_executor(
            self._executor,
            partial(self._client.multi_query, *args, **kwargs),
        )

    async def query_bulk(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> BulkOnlineQueryResponse:
        return await self._current_loop().run_in_executor(
            self._executor,
            partial(self._client.query_bulk, *args, **kwargs),
        )

    async def offline_query(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> DatasetImpl:
        return await self._current_loop().run_in_executor(
            self._executor,
            partial(self._client.offline_query, *args, **kwargs),
        )

    async def create_dataset(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> DatasetImpl:
        return await self._current_loop().run_in_executor(
            self._executor,
            partial(self._client.create_dataset, *args, **kwargs),
        )

    async def prompt_evaluation(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Dataset:
        return await self._current_loop().run_in_executor(
            self._executor,
            partial(self._client.prompt_evaluation, *args, **kwargs),
        )

    async def upload_features(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> UploadFeaturesResponse:
        errors: list[ChalkError] = await self._current_loop().run_in_executor(
            self._executor, partial(self._client.upload_features, *args, **kwargs)
        )
        return UploadFeaturesResponse(errors=errors)

    async def get_operation_feature_statistics(self, operation_id: UUID) -> FeatureStatisticsResponse:
        return await self._current_loop().run_in_executor(
            self._executor,
            partial(self._client.get_operation_feature_statistics, operation_id),
        )


ASYNC_CHALK_CLIENT_EXECUTOR = ChalkThreadPoolExecutor(
    max_workers=int(os.getenv("CHALK_IO_EXECUTOR_MAX_WORKERS", "32")),
    thread_name_prefix="async-chalk-client-",
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: chalk/auth/v1/permissions.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from chalk._gen.chalk.utils.v1 import encoding_pb2 as chalk_dot_utils_dot_v1_dot_encoding__pb2
from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b"\n\x1f\x63halk/auth/v1/permissions.proto\x12\rchalk.auth.v1\x1a\x1d\x63halk/utils/v1/encoding.proto\x1a google/protobuf/descriptor.proto*\xf6\n\n\nPermission\x12\x1a\n\x16PERMISSION_UNSPECIFIED\x10\x00\x12'\n#PERMISSION_INSECURE_UNAUTHENTICATED\x10\x01\x12\x1c\n\x18PERMISSION_AUTHENTICATED\x10\x02\x12\x1b\n\x17PERMISSION_QUERY_ONLINE\x10\x03\x12\x1c\n\x18PERMISSION_QUERY_OFFLINE\x10\x04\x12 \n\x1cPERMISSION_MONITORING_CREATE\x10\x05\x12\x1e\n\x1aPERMISSION_MONITORING_READ\x10\x06\x12\x17\n\x13PERMISSION_TEAM_ADD\x10\x07\x12\x1a\n\x16PERMISSION_TEAM_DELETE\x10\x08\x12\x18\n\x14PERMISSION_TEAM_LIST\x10\t\x12\x19\n\x15PERMISSION_TEAM_ADMIN\x10\n\x12\x1a\n\x16PERMISSION_DEPLOY_READ\x10\x0b\x12\x1c\n\x18PERMISSION_DEPLOY_CREATE\x10\x0c\x12\x1d\n\x19PERMISSION_DEPLOY_PREVIEW\x10\r\x12\x1e\n\x1aPERMISSION_DEPLOY_REDEPLOY\x10\x0e\x12\x18\n\x14PERMISSION_LOGS_LIST\x10\x0f\x12\x18\n\x14PERMISSION_CRON_READ\x10\x10\x12\x1a\n\x16PERMISSION_CRON_CREATE\x10\x11\x12\x1c\n\x18PERMISSION_SECRETS_WRITE\x10\x12\x12\x1e\n\x1aPERMISSION_SECRETS_DECRYPT\x10\x13\x12\x1b\n\x17PERMISSION_SECRETS_LIST\x10\x14\x12\x1b\n\x17PERMISSION_TOKENS_WRITE\x10\x15\x12\x1a\n\x16PERMISSION_TOKENS_LIST\x10\x16\x12\x1b\n\x17PERMISSION_MIGRATE_READ\x10\x17\x12\x1b\n\x17PERMISSION_MIGRATE_PLAN\x10\x18\x12\x1e\n\x1aPERMISSION_MIGRATE_EXECUTE\x10\x19\x12\x1d\n\x19PERMISSION_PROJECT_CREATE\x10\x1a\x12\x1a\n\x16PERMISSION_CHALK_ADMIN\x10\x1b\x12\x1b\n\x17PERMISSION_BILLING_READ\x10\x1c\x1a\x8e\x04\xe2\xa1'\x89\x04\n\x1c\x08\x01\x12\x18insecure_unauthenticated\n\x11\x08\x02\x12\rauthenticated\n\x10\x08\x03\x12\x0cquery.online\n\x11\x08\x04\x12\rquery.offline\n\x15\x08\x05\x12\x11monitoring.create\n\x13\x08\x06\x12\x0fmonitoring.read\n\x0c\x08\x07\x12\x08team.add\n\x0f\x08\x08\x12\x0bteam.delete\n\r\x08\t\x12\tteam.list\n\x0e\x08\n\x12\nteam.admin\n\x0f\x08\x0b\x12\x0b\x64\x65ploy.read\n\x11\x08\x0c\x12\rdeploy.create\n\x12\x08\r\x12\x0e\x64\x65ploy.preview\n\x13\x08\x0e\x12\x0f\x64\x65ploy.redeploy\n\r\x08\x0f\x12\tlogs.list\n\r\x08\x10\x12\tcron.read\n\x0f\x08\x11\x12\x0b\x63ron.create\n\x11\x08\x12\x12\rsecrets.write\n\x13\x08\x13\x12\x0fsecrets.decrypt\n\x10\x08\x14\x12\x0csecrets.list\n\x10\x08\x15\x12\x0ctokens.write\n\x0f\x08\x16\x12\x0btokens.list\n\x10\x08\x17\x12\x0cmigrate.read\n\x10\x08\x18\x12\x0cmigrate.plan\n\x13\x08\x19\x12\x0fmigrate.execute\n\x12\x08\x1a\x12\x0eproject.create\n\x0f\x08\x1b\x12\x0b\x63halk.admin\n\x10\x08\x1c\x12\x0c\x62illing.read:Z\n\npermission\x12\x1e.google.protobuf.MethodOptions\x18\xd0\x0f \x01(\x0e\x32\x19.chalk.auth.v1.PermissionR\npermissionB\x8b\x01\n\x11\x63om.chalk.auth.v1B\x10PermissionsProtoP\x01Z\x0e\x61uth/v1;authv1\xa2\x02\x03\x43\x41X\xaa\x02\rChalk.Auth.V1\xca\x02\rChalk\\Auth\\V1\xe2\x02\x19\x43halk\\Auth\\V1\\GPBMetadata\xea\x02\x0f\x43halk::Auth::V1b\x06proto3"
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "chalk.auth.v1.permissions_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals[
        "DESCRIPTOR"
    ]._serialized_options = b"\n\021com.chalk.auth.v1B\020PermissionsProtoP\001Z\016auth/v1;authv1\242\002\003CAX\252\002\rChalk.Auth.V1\312\002\rChalk\\Auth\\V1\342\002\031Chalk\\Auth\\V1\\GPBMetadata\352\002\017Chalk::Auth::V1"
    _globals["_PERMISSION"]._options = None
    _globals[
        "_PERMISSION"
    ]._serialized_options = b"\342\241'\211\004\n\034\010\001\022\030insecure_unauthenticated\n\021\010\002\022\rauthenticated\n\020\010\003\022\014query.online\n\021\010\004\022\rquery.offline\n\025\010\005\022\021monitoring.create\n\023\010\006\022\017monitoring.read\n\014\010\007\022\010team.add\n\017\010\010\022\013team.delete\n\r\010\t\022\tteam.list\n\016\010\n\022\nteam.admin\n\017\010\013\022\013deploy.read\n\021\010\014\022\rdeploy.create\n\022\010\r\022\016deploy.preview\n\023\010\016\022\017deploy.redeploy\n\r\010\017\022\tlogs.list\n\r\010\020\022\tcron.read\n\017\010\021\022\013cron.create\n\021\010\022\022\rsecrets.write\n\023\010\023\022\017secrets.decrypt\n\020\010\024\022\014secrets.list\n\020\010\025\022\014tokens.write\n\017\010\026\022\013tokens.list\n\020\010\027\022\014migrate.read\n\020\010\030\022\014migrate.plan\n\023\010\031\022\017migrate.execute\n\022\010\032\022\016project.create\n\017\010\033\022\013chalk.admin\n\020\010\034\022\014billing.read"
    _globals["_PERMISSION"]._serialized_start = 116
    _globals["_PERMISSION"]._serialized_end = 1514
# @@protoc_insertion_point(module_scope)
